"""### Functions that interact with Conda via shell."""

import os
from typing import Optional

def get_conda_prefix_path() -> Optional[str]:
  """
  ### Returns the path for the current Conda enviroment.

  #### Returns:
  - (str | None): Path for current Conda enviroment.
  """
  return os.environ.get('CONDA_PREFIX')



"""
### Mode Sync Executor Module.

This module contains the base class for mode executors with sync operations.
"""
import os
from abc import abstractmethod
from typing import Dict, Tuple

from xmipp3_installer.application.logger import errors
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes import mode_executor

_SYNC_PROGRAM_PATH = os.path.join(".", paths.BINARIES_PATH)
_SYNC_PROGRAM_NAME = "xmipp_sync_data"

class ModeSyncExecutor(mode_executor.ModeExecutor):
  """Base class for mode executors with sync operations."""

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.sync_program_name = _SYNC_PROGRAM_NAME
    self.sync_program_path = os.path.join(_SYNC_PROGRAM_PATH, _SYNC_PROGRAM_NAME)

  def run(self) -> Tuple[int, str]:
    """
    ### Executes the sync operation.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    if not os.path.exists(self.sync_program_path):
      logger('\n'.join([
        logger.red(f"{self.sync_program_path} does not exist."),
        logger.red("Xmipp needs to be compiled successfully before running this command!")
      ]))
      return errors.IO_ERROR, ""
    return self._sync_operation()

  @abstractmethod
  def _sync_operation(self) -> Tuple[int, str]:
    """
    ### Executes the specific sync operation.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """



"""
### Mode Config Executor Module.

This module contains the class to manage the configuration file.
"""

import os
from typing import Dict, Tuple

from xmipp3_installer.application.logger import errors
from xmipp3_installer.application.logger import predefined_messages
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes import mode_executor
from xmipp3_installer.repository import config

class ModeConfigExecutor(mode_executor.ModeExecutor):
  """
  ### Mode Config Executor.

  Manages the configuration file for the installation.
  """

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.overwrite = context.pop(params.PARAM_OVERWRITE)
    self.config_values = {}
  
  def run(self) -> Tuple[int, str]:
    """
    ### Reads the config file and writes to it formatting properly with the appropiate values.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error. 
    """
    logger(predefined_messages.get_section_message("Managing config file"))
    action_message = (
      "Generating config file from scratch with default values..."
      if self.overwrite or not os.path.exists(paths.CONFIG_FILE) else
      "Reading config file..."
    )
    logger(action_message)
    try:
      file_handler = config.ConfigurationFileHandler()
      file_handler.write_config(overwrite=self.overwrite)
    except PermissionError as permission_error:
      return errors.IO_ERROR, str(permission_error)
    self.config_values = file_handler.values
    logger(predefined_messages.get_done_message())
    return 0, ""



"""### Contains all constants needed for handling errors during Xmipp's installation."""

from xmipp3_installer.installer.urls import CMAKE_INSTALL_DOCS_URL
from xmipp3_installer.installer.constants import paths

# Error codes
INTERRUPTED_ERROR = -1
OK = 0
UNKOW_ERROR = 1
SOURCE_CLONE_ERROR = 2
CMAKE_ERROR = 3
CMAKE_CONFIGURE_ERROR = 4
CMAKE_COMPILE_ERROR = 5
CMAKE_INSTALL_ERROR = 6
IO_ERROR = 7

# Error messages
__CHECK_LOG_MESSAGE = f'Check the inside file \'{paths.LOG_FILE}\'.'
ERROR_CODES = {
  INTERRUPTED_ERROR: ['Process was interrupted by the user.', ''],
  UNKOW_ERROR: ['', ''],
  SOURCE_CLONE_ERROR: ['Error cloning xmipp repository with git.', 'Please review the internet connection and the git package.'],
  CMAKE_ERROR: ['There was an error with CMake.', f'Please install it by following the instructions at {CMAKE_INSTALL_DOCS_URL}'],
  CMAKE_CONFIGURE_ERROR: ['Error configuring with CMake.', __CHECK_LOG_MESSAGE],
  CMAKE_COMPILE_ERROR: ['Error compiling with CMake.', __CHECK_LOG_MESSAGE],
  CMAKE_INSTALL_ERROR: ['Error installing with CMake.', __CHECK_LOG_MESSAGE],
  IO_ERROR: ['Input/output error.', 'This error can be caused by the installer not being able to read/write/create/delete a file. Check your permissions on this directory.']
}



"""### Contains the configuration file singleton that interact with the configuration file."""

import re
import os
from datetime import datetime
from typing import List, Optional, Tuple, Dict, Any

from xmipp3_installer.shared.singleton import Singleton
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.repository.config_vars import default_values, variables, config_values_adapter
from xmipp3_installer.repository.invalid_config_line import InvalidConfigLineError

class ConfigurationFileHandler(Singleton):
  """### Configuration file class for loading and storing the installation configuration."""

  __COMMENT_ESCAPE = '#'
  __ASSIGNMENT_SEPARATOR = '='
  __LAST_MODIFIED_TEXT = "Config file automatically generated on"

  def __init__(self, path: str=paths.CONFIG_FILE, show_errors: bool=True):
    """
    ### Constructor.
    
    #### Params:
    - path (str): Optional. Path to the configuration file.
    """
    self.__path = path
    self.show_errors = show_errors
    self.values = {}
    self.read_config()
    self.last_modified = self.__read_config_date()

  def read_config(self):
    """### Reads the config file and stores a dictionary with all the parsed variables."""
    file_lines = self.__get_file_content()
    result = {}
    for line_number, line in enumerate(file_lines):
      new_result = self.__add_line_values(result.copy(), line, line_number + 1)
      if new_result is None:
        result = {}
        break
      result = new_result
    self.values = {
      **config_values_adapter.get_context_values_from_file_values(
        default_values.CONFIG_DEFAULT_VALUES
      ),
      **config_values_adapter.get_context_values_from_file_values(
        result,
        show_warnings=self.show_errors
      )
    }

  def write_config(self, overwrite: bool=False):
    """
    ### Writes a template config file with stored variables, leaving the rest with default values.

    #### Params:
    - overwrite (bool): Optional. If True, default values are written, overwriting the existing file.
    """
    if overwrite:
      self.values = config_values_adapter.get_context_values_from_file_values(default_values.CONFIG_DEFAULT_VALUES)
    values = config_values_adapter.get_file_values_from_context_values(self.values.copy())
    self.last_modified = datetime.today().strftime('%d-%m-%Y %H:%M.%S')
    
    lines = ["##### TOGGLE SECTION #####\n"]
    lines.append(f"# Activate or deactivate this features using values {default_values.ON}/{default_values.OFF}\n")
    lines.extend(self.__get_section_lines(variables.TOGGLES, values))

    lines.append("\n##### PACKAGE HOME SECTION #####\n")
    lines.append("# Use this variables to use custom installation paths for the required packages.\n")
    lines.append("# If left empty, CMake will search for those packages within your system.\n")
    lines.extend(self.__get_section_lines(variables.LOCATIONS, values))
    
    lines.append("\n##### COMPILATION FLAGS #####\n")
    lines.append("# We recommend not modifying this variables unless you know what you are doing.\n")
    lines.extend(self.__get_section_lines(variables.COMPILATION_FLAGS, values))
    
    if values:
      lines.append("\n##### UNKNOWN VARIABLES #####\n")
      lines.append("# This variables were not expected, but are kept here in case they might be needed.\n")
      lines.extend(self.__get_unkown_variable_lines(values))

    lines.append(f"\n# {self.__LAST_MODIFIED_TEXT} {self.last_modified}\n")
    with open(self.__path, 'w', encoding="utf-8") as config_file:
      config_file.writelines(lines)

  def get_config_date(self) -> str:
    """
    ### Returns the date of the last modification of the configuration file.

    #### Returns:
    - (str): Date in dd-mm-yyyy format.
    """
    if not self.last_modified:
      self.last_modified = self.__read_config_date()
    return self.last_modified
    
  def __get_file_content(self) -> List[str]:
    """
    ### Reads the whole unparsed content of the given file.

    #### Returns:
    - (list(str)): Content of the file, where each line is a string in the result list.
    """
    if not os.path.exists(self.__path):
      return []
    lines = []
    with open(self.__path, encoding="utf-8") as config_file:
      lines = config_file.readlines()
    return lines

  def __read_config_date(self) -> str:
    """
    ### Reads from the config file the date of its last modification.

    #### Returns:
    - (str): Date in dd-mm-yyyy format.
    """
    config_lines = self.__get_file_content()
    for line in config_lines:
      if self.__LAST_MODIFIED_TEXT not in line:
        continue
      match = re.search(r'\d{2}-\d{2}-\d{4} \d{2}:\d{2}.\d{2}', line)
      if match:
        return match.group()
    return ""

  def __add_line_values(self, config: Dict, line: str, line_number: int) -> Optional[Dict]:
    """
    ### Adds the config values present in the current line to the given dictionary.

    #### Params:
    - config (dict): Dictionary with all the present variables.
    - line (str): Current line to extract values from.
    - line_number (int): Line index in the list of lines that compose the file.

    #### Returns:
    - (dict | None): An updated dictionary with the newly obtained variable, or None if there was an error.
    """
    try:
      key_value_pair = self.__parse_config_line(line, line_number)
    except InvalidConfigLineError as error:
      if self.show_errors:
        logger(str(error))
      config = {}
      return
    if key_value_pair:
      key, value = key_value_pair
      config[key] = value
    return config

  def __parse_config_line(self, line: str, line_number: int) -> Optional[Tuple[str, str]]:
    """
    ### Reads the given line from the config file and returns the key-value pair as a tuple.

    #### Params:
    - line_number (int): Line number inside the config file.
    - line (str): Line to parse.
    
    #### Returns:
    - (tuple(str, str) | None): Tuple containing the read key-value pair if line contains valid data.

    #### Raises:
    - RuntimeError: Raised when a line has an invalid format and cannot be parsed.
    """
    N_TOKENS = 2
    line_without_comments = line.split(self.__COMMENT_ESCAPE, maxsplit=2)[0].strip()
    if not line_without_comments:
      return None
    
    tokens = line_without_comments.split(self.__ASSIGNMENT_SEPARATOR, maxsplit=1)
    if len(tokens) != N_TOKENS:
      raise InvalidConfigLineError(
        InvalidConfigLineError.generate_error_message(
          paths.CONFIG_FILE,
          line_number,
          line
        )
      )
    
    return tokens[0].strip(), tokens[1].strip()

  def __make_config_line(self, key: str, value: Optional[str], default_value: str) -> str:
    """
    ### Composes a config file line given a key-value pair to write.

    #### Params:
    - key (int): Name of the variable.
    - value (str | None): Value of the variable found in the config file.
    - default_value (str): Default value of the variable.
    
    #### Returns:
    - (str): String containing the appropiately formatted key-value pair.
    """
    default_value = '' if default_value is None else default_value
    value = default_value if value is None else value
    return f"{key}{self.__ASSIGNMENT_SEPARATOR}{value}" if key else ""

  def __get_section_lines(self, section_type: str, config_variables: Dict[str, str]) -> List[str]:
    """
    ### Returns the lines composed by the given section's variables in the dictionary, and deletes them from it.

    #### Params:
    - section_type (str): Section to extract variables from.
    - config_variables (dict(str, str)): Dictionary containing all variables.

    #### Returns:
    - (list(str)): Config file lines created from the dictionary variables.
    """
    lines = []
    for section_variable in variables.CONFIG_VARIABLES[section_type]:
      lines.append(''.join([
        self.__make_config_line(
          section_variable,
          config_variables.get(section_variable),
          default_values.CONFIG_DEFAULT_VALUES[section_variable]
        ),
        "\n"
      ]))
      config_variables.pop(section_variable, None)
    return lines

  def __get_unkown_variable_lines(self, config_variables: Dict[str, Any]) -> List[str]:
    """
    ### Returns the lines composed by the unkown variables in the dictionary.

    #### Params:
    - config_variables (dict(str, any)): Dictionary containing all unknown variables.

    #### Returns:
    - (list(str)): Config file lines created from the dictionary variables.
    """
    lines = []
    for variable_name, value in config_variables.items():
      lines.append(
        f"{self.__make_config_line(variable_name, value, '')}\n"
      )
    return lines



"""
### User Interactions Module.

This module contains functions to handle user interactions.
"""

def get_user_confirmation(confirmation_text: str, case_sensitive: bool=True) -> bool:
  """
  ### Receives confirmation from user input.

  #### Params:
  - confirmation_text (str): The text the user needs to input to confirm.
  - case_sensitive (bool): Optional. If True, the text entered has to exactly match.

  #### Returns:
  - (bool): True if the user confirmed, False otherwise.
  """
  received_input = input()
  if not case_sensitive:
    received_input = received_input.lower()
    confirmation_text = confirmation_text.lower()
  return received_input == confirmation_text


"""### Functions that interact with the shell."""

import os
import subprocess
import threading

from typing import Tuple

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.application.logger import errors

def run_shell_command(
  cmd: str,
  *,
  cwd: str='./',
  show_command: bool=False,
  show_output: bool=False,
  show_error: bool=False
) -> Tuple[int, str]:
  """
  ### This function runs the given command.

  #### Params:
  - cmd (str): Command to run.
  - cwd (str): Optional. Path to run the command from. Default is current directory.
  - show_output (bool): Optional. If True, output is printed.
  - show_error (bool): Optional. If True, errors are printed.
  - show_command (bool): Optional. If True, command is printed in blue.

  #### Returns:
  - (int): Return code.
  - (str): Output of the command, regardless of if it is an error or regular output.
  """
  if show_command:
    logger(logger.blue(cmd))
  ret_code, output_str = __run_command(cmd, cwd=cwd)

  if not ret_code and show_output:
    logger(output_str)
  
  if ret_code and show_error:
    logger(logger.red(output_str))
  
  return ret_code, output_str

def run_shell_command_in_streaming(
  cmd: str,
  cwd: str='./',
  show_output: bool=False,
  show_error: bool=False,
  substitute: bool=False
) -> int:
  """
  ### Runs the given command and shows its output as it is being generated.

  #### Params:
  - cmd (str): Command to run.
  - cwd (str): Optional. Path to run the command from. Default is current directory.
  - show_output (bool): Optional. If True, output is printed.
  - show_error (bool): Optional. If True, errors are printed.
  - substitute (bool): Optional. If True, output will replace previous line.

  #### Returns:
  - (int): Return code.
  """
  logger(cmd, substitute=substitute)
  process = subprocess.Popen(cmd, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
  
  thread_out = threading.Thread(
    target=logger.log_in_streaming,
    args=(process.stdout,),
    kwargs={"show_in_terminal": show_output, "substitute": substitute, "err": False}
  )
  thread_err = threading.Thread(
    target=logger.log_in_streaming,
    args=(process.stderr,),
    kwargs={"show_in_terminal": show_error, "substitute": substitute, "err": True}
  )
  thread_out.start()
  thread_err.start()

  try:
    process.wait()
    thread_out.join()
    thread_err.join()
  except KeyboardInterrupt:
    process.returncode = errors.INTERRUPTED_ERROR
  
  return process.returncode

def __run_command(cmd: str, cwd: str='./') -> Tuple[int, str]:
  """
  ### Runs the given shell command.

  #### Params:
  - cmd (str): Command to run.
  - cwd (str): Optional. Path to run the command from.

  #### Returns:
  - (int): Return code of the operation.
  - (str): Return message of the operation.
  """
  process = subprocess.Popen(cmd, cwd=cwd, env=os.environ, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
  try:
    process.wait()
  except KeyboardInterrupt:
    return errors.INTERRUPTED_ERROR, ""
  
  ret_code = process.returncode
  output, err = process.communicate()
  output_str = output.decode() if not ret_code and output else err.decode()
  output_str = output_str[:-1] if output_str.endswith('\n') else output_str
  return ret_code, output_str



"""### Contains the default values for the config variables."""

from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import conda_handler
from xmipp3_installer.repository.config_vars import variables

__TUNE_FLAG = '-mtune=native'

ON = 'ON'
OFF = 'OFF'
CONFIG_DEFAULT_VALUES = {
  variables.SEND_INSTALLATION_STATISTICS: ON,
  variables.CMAKE: None,
  variables.CUDA: ON,
  variables.MPI: ON,
  variables.CC: None,
  variables.CXX: None,
  variables.CMAKE_INSTALL_PREFIX: paths.INSTALL_PATH,
  variables.CC_FLAGS: __TUNE_FLAG,
  variables.CXX_FLAGS: __TUNE_FLAG,
  variables.CUDA_COMPILER: None,
  variables.PREFIX_PATH: conda_handler.get_conda_prefix_path(),
  variables.MPI_HOME: None,
  variables.PYTHON_HOME: None,
  variables.FFTW_HOME: None,
  variables.TIFF_HOME: None,
  variables.HDF5_HOME: None,
  variables.JPEG_HOME: None,
  variables.SQLITE_HOME: None,
  variables.CUDA_CXX: None,
  variables.MATLAB: ON,
  variables.LINK_SCIPION: ON,
  variables.BUILD_TESTING: OFF,
  variables.SKIP_RPATH: ON,
  variables.BUILD_TYPE: "Release"
}



"""
### Mode Version Executor Module.

This module contains the class to collect and display version information.
"""
import os
from typing import Dict, Tuple

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.api_client.assembler import installation_info_assembler
from xmipp3_installer.installer import constants
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes import mode_executor
from xmipp3_installer.installer.handlers import git_handler, versions_manager
from xmipp3_installer.installer.handlers.cmake import cmake_handler
from xmipp3_installer.repository.config_vars import variables

class ModeVersionExecutor(mode_executor.ModeExecutor):
  """
  ### Mode Version Executor.

  Collects and displays version information for the installation.
  """
  
  __LEFT_TEXT_LEN = 25

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.short = context.pop(params.PARAM_SHORT)
    config_exists = os.path.exists(paths.CONFIG_FILE)
    self.version_file_exists = os.path.exists(paths.LIBRARY_VERSIONS_FILE)
    self.is_configured = config_exists and self.version_file_exists
    versions: versions_manager.VersionsManager = context[constants.VERSIONS_CONTEXT_KEY]
    self.xmipp_version_number = versions.xmipp_version_number
    self.xmipp_version_name = versions.xmipp_version_name
    self.release_date = versions.xmipp_release_date
  
  def run(self) -> Tuple[int, str]:
    """
    ### Collects all the version information available and displays it.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error.
    """
    installation_info =  self.xmipp_version_name if self.short else self.__get_long_version()
    logger(installation_info)
    return 0, ""

  def __get_long_version(self) -> str:
    """
    ### Returns the long version of the installation info.

    #### Returns:
    - (str): Long version of the installation info.
    """
    installation_info_lines = []
    version_type = 'release' if git_handler.is_tag() else git_handler.get_current_branch()
    title = f"Xmipp {self.xmipp_version_number} ({version_type})"
    installation_info_lines.append(f"{logger.bold(title)}\n")
    installation_info_lines.append(self.__get_dates_section())
    system_version_left_text = self.__add_padding_spaces("System version: ")
    installation_info_lines.append(f"{system_version_left_text}{installation_info_assembler.get_os_release_name()}")
    installation_info_lines.append(self.__get_sources_info())
    if self.version_file_exists:
      installation_info_lines.append(f"\n{self.__get_library_versions_section()}")
    if not self.is_configured or not _are_all_sources_present():
      installation_info_lines.append(f"\n{_get_configuration_warning_message()}")
    return '\n'.join(installation_info_lines)

  def __get_dates_section(self) -> str:
    """
    ### Returns the message section related to dates.

    #### Returns:
    - (str): Dates related message section.
    """
    dates_section = f"{self.__add_padding_spaces('Release date: ')}{self.release_date}\n"
    dates_section += f"{self.__add_padding_spaces('Compilation date: ')}"
    last_modified = self.context.get(variables.LAST_MODIFIED_KEY)
    dates_section += last_modified if last_modified else '-'
    return dates_section

  def __get_sources_info(self) -> str:
    """
    ### Returns the message section related to sources.

    #### Returns:
    - (str): Sources related message section.
    """
    sources_message_lines = []
    for source_package in constants.XMIPP_SOURCES:
      sources_message_lines.append(self.__get_source_info(source_package))
    return '\n'.join(sources_message_lines)

  def __get_source_info(self, source: str) -> str:
    """
    ### Returns the info message related to a given source.

    #### Params:
    - source (str): Source to get the message about.

    #### Returns:
    - (str): Info message about the given source.
    """
    source_path = paths.get_source_path(source)
    source_left_text = self.__add_padding_spaces(f"{source} branch: ")
    if not os.path.exists(source_path):
      return f"{source_left_text}{logger.yellow('Not found')}"
    current_commit = git_handler.get_current_commit(dir=source_path)
    commit_branch = git_handler.get_commit_branch(current_commit, dir=source_path)
    current_branch = git_handler.get_current_branch(dir=source_path)
    display_name = commit_branch if git_handler.is_tag(dir=source_path) else current_branch
    return f"{source_left_text}{display_name} ({current_commit})"

  def __add_padding_spaces(self, left_text: str) -> str:
    """
    ### Adds right padding as spaces to the given text until it reaches the desired length.

    #### Params:
    - left_text (str): Text to add padding to.

    #### Returns:
    - (str): Padded string.
    """
    text_len = len(left_text)
    if text_len >= self.__LEFT_TEXT_LEN:
      return left_text
    spaces = ''.join([' ' for _ in range(self.__LEFT_TEXT_LEN - text_len)])
    return f"{left_text}{spaces}"

  def __get_library_versions_section(self) -> str:
    """
    ### Retrieves the version of the libraries used in the project.
    
    #### Returns:
    - (str): Libraries with their version.
    """
    version_lines = [] 
    for library, version in cmake_handler.get_library_versions_from_cmake_file(
      paths.LIBRARY_VERSIONS_FILE
    ).items():
      library_left_text = self.__add_padding_spaces(f"{library}: ")
      version_lines.append(f"{library_left_text}{version}")
    return '\n'.join(version_lines)
  
def _are_all_sources_present() -> bool:
  """
  ### Check if all required source packages are present.

  #### Returns:
  - (bool): True if all source packages are present, False otherwise.
  """
  for source_package in paths.XMIPP_SOURCE_PATHS:
    if not os.path.exists(source_package):
      return False
  return True

def _get_configuration_warning_message() -> str:
  """
  ### Returns a message indicating configuration is not complete.

  #### Returns:
  - (str): 
  """
  return '\n'.join([
    logger.yellow("This project has not yet been configured, so some detectable dependencies have not been properly detected."),
    logger.yellow("Run mode 'getSources' and then 'configBuild' to be able to show all detectable ones.")
  ])



"""
### Mode Test Executor Module.

This module contains the class to execute Xmipp tests.
"""
import os
from typing import Dict, Tuple, Any

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger.logger import logger, errors
from xmipp3_installer.installer import urls
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler
from xmipp3_installer.installer.modes.mode_sync.mode_sync_executor import ModeSyncExecutor
from xmipp3_installer.repository.config_vars import variables

_DATASET_NAME = "xmipp_programs"
_PYTHON_TEST_SCRIPT_PATH = os.path.join(paths.SOURCES_PATH, "xmipp")
_PYTHON_TEST_SCRIPT_INTERNAL_FOLDER = "tests"
_PYTHON_TEST_SCRIPT_NAME = "test.py"
_PYTHON_DATA_FOLDER = 'data'
_PYTHON_TEST_SCRIPT_INTERNAL_PATH = os.path.join(
  _PYTHON_TEST_SCRIPT_INTERNAL_FOLDER,
  _PYTHON_TEST_SCRIPT_NAME
)
_DEFAULT_PYTHON_HOME = "python3"
_DATASET_PATH = os.path.join(_PYTHON_TEST_SCRIPT_PATH, _PYTHON_TEST_SCRIPT_INTERNAL_FOLDER,  _PYTHON_DATA_FOLDER)
_TEST_DATA = os.path.join(_PYTHON_TEST_SCRIPT_INTERNAL_FOLDER, _PYTHON_DATA_FOLDER)
_BASHRC_FILE_PATH = os.path.abspath(os.path.join(paths.INSTALL_PATH, "xmipp.bashrc"))
_TESTS_SEPARATOR = " "
_PARAM_MAPPER = {
  params.PARAM_SHOW_TESTS: "--show",
  params.PARAM_ALL_FUNCTIONS: "--allFuncs",
  params.PARAM_ALL_PROGRAMS: "--allPrograms"
}

class ModeTestExecutor(ModeSyncExecutor):
  """Class to execute Xmipp tests."""

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.param_value = self.__get_selected_param_value(context)
    self.cuda = context.pop(variables.CUDA)
    python_home = context.pop(variables.PYTHON_HOME, None)
    self.python_home = python_home if python_home else _DEFAULT_PYTHON_HOME
  
  def run(self) -> Tuple[int, str]:
    """
    ### Runs the provided tests.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    ret_code, output = self.__load_bashrc()
    if ret_code:
      return ret_code, output
    
    ret_code, output = super().run()
    if ret_code:
      return ret_code, output
    return self.__run_tests()

  def _sync_operation(self) -> Tuple[int, str]:
    """
    ### Executes the test operation.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    if os.path.isdir(_DATASET_PATH):
      task_message = "Updating"
      task = "update"
    else:
      task_message = "Downloading"
      task = "download"
    logger(logger.blue(f"{task_message} the test files"))

    args = f"{_TEST_DATA} {urls.SCIPION_TESTS_URL} {_DATASET_NAME}"
    ret_code = shell_handler.run_shell_command_in_streaming(
      f"{self.sync_program_name} {task} {args}",
      cwd=_PYTHON_TEST_SCRIPT_PATH,
      show_output=True,
      show_error=True,
      substitute=True
    )
    ret_code = 1 if ret_code else ret_code
    return ret_code, ""

  @staticmethod
  def __load_bashrc() -> Tuple[int, str]:
    """
    ### Loads the bashrc file.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    if not os.path.exists(_BASHRC_FILE_PATH):
      return errors.IO_ERROR, f"File {_BASHRC_FILE_PATH} does not exist."

    ret_code, output = shell_handler.run_shell_command(
      f"bash -c 'source {_BASHRC_FILE_PATH} && env'"
    )
    if ret_code:
      return 1, output

    for line in output.splitlines():
      key, _, value = line.partition("=")
      os.environ[key] = value
    return 0, ""

  def __run_tests(self) -> Tuple[int, str]:
    """
    ### Runs the specified tests.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    no_cuda_str = "--noCuda" if not self.cuda else ""
    if self.param_value not in {
      _PARAM_MAPPER[params.PARAM_SHOW_TESTS],
      _PARAM_MAPPER[params.PARAM_ALL_FUNCTIONS],
      _PARAM_MAPPER[params.PARAM_ALL_PROGRAMS]
    }:
      test_names = self.param_value.replace(_TESTS_SEPARATOR, ", ")
      logger(f" Tests to run: {test_names}")
    ret_code, output = shell_handler.run_shell_command(
      f"{self.python_home} {_PYTHON_TEST_SCRIPT_INTERNAL_PATH} {self.param_value} {no_cuda_str}",
      cwd=_PYTHON_TEST_SCRIPT_PATH,
      show_output=True
    )
    ret_code = 1 if ret_code else 0
    return ret_code, output

  @staticmethod
  def __get_selected_param_value(context: Dict[str, Any]) -> str:
    """
    ### Returns the value of the param selected to run the test execution.

    #### Params:
    - context (dict(str, any)): Dictionary containing the installation context variables.

    #### Returns:
    - (str): The value of the selected variable to use
    """
    for variable in {
      params.PARAM_SHOW_TESTS,
      params.PARAM_ALL_FUNCTIONS,
      params.PARAM_ALL_PROGRAMS
    }:
      if context[variable]:
        return _PARAM_MAPPER[variable]
    return _TESTS_SEPARATOR.join(context.pop(params.PARAM_TEST_NAMES))



"""### Containis all param constants needed for the argument parsing part of Xmipp's installation."""

from xmipp3_installer.application.cli import arguments

# Definition of all params found in the
SHORT_VERSION = 'short'
LONG_VERSION = 'long'
DESCRIPTION = 'description'

# Possible param list
PARAM_SHORT = 'short'
PARAM_JOBS = 'jobs'
PARAM_BRANCH = 'branch'
PARAM_MODELS_DIRECTORY = 'directory'
PARAM_TEST_NAMES = 'testNames'
PARAM_SHOW_TESTS = 'show'
PARAM_ALL_FUNCTIONS = 'all_functions'
PARAM_ALL_PROGRAMS = 'all_programs'
PARAM_GIT_COMMAND = 'command'
PARAM_LOGIN = 'login'
PARAM_MODEL_PATH = 'modelPath'
PARAM_UPDATE = 'update'
PARAM_OVERWRITE = 'overwrite'
PARAM_KEEP_OUTPUT = "keep_output"
PARAMS = {
  PARAM_SHORT: {
    LONG_VERSION: "--short",
    DESCRIPTION: "If set, only version number is shown."
  },
  PARAM_JOBS: {
    SHORT_VERSION: "-j",
    LONG_VERSION: "--jobs",
    DESCRIPTION: "Number of jobs. Defaults to all available."
  },
  PARAM_BRANCH: {
    SHORT_VERSION: "-b",
    LONG_VERSION: "--branch",
    DESCRIPTION: "Branch for the source repositories."
  },
  PARAM_MODELS_DIRECTORY: {
    SHORT_VERSION: "-d",
    LONG_VERSION: "--directory",
    DESCRIPTION: f"Directory where models will be saved. Default is \"{arguments.DEFAULT_MODELS_DIR}\"."
  },
  PARAM_TEST_NAMES: {
    SHORT_VERSION: PARAM_TEST_NAMES,
    DESCRIPTION: "Name of the tests to run."
  },
  PARAM_SHOW_TESTS: {
    LONG_VERSION: "--show",
    DESCRIPTION: "Shows the tests available and how to invoke those."
  },
  PARAM_ALL_FUNCTIONS : {
    LONG_VERSION: "--all-functions",
    DESCRIPTION: "If set, all function tests will be run."
  },
  PARAM_ALL_PROGRAMS : {
    LONG_VERSION: "--all-programs",
    DESCRIPTION: "If set, all program tests will be run."
  },
  PARAM_GIT_COMMAND: {
    SHORT_VERSION: PARAM_GIT_COMMAND,
    DESCRIPTION: "Git command to run on all source repositories."
  },
  PARAM_LOGIN: {
    SHORT_VERSION: "login",
    DESCRIPTION: "Login (usr@server) for remote host to upload the model with. Must have write permissions to such machine."
  },
  PARAM_MODEL_PATH: {
    SHORT_VERSION: PARAM_MODEL_PATH,
    DESCRIPTION: "Path to the model to upload to remote host."
  },
  PARAM_UPDATE: {
    LONG_VERSION: "--update",
    DESCRIPTION: "Flag to update an existing model"
  },
  PARAM_OVERWRITE: {
    SHORT_VERSION: "-o",
    LONG_VERSION: "--overwrite",
    DESCRIPTION: "If set, current config file will be overwritten with a new one."
  },
  PARAM_KEEP_OUTPUT: {
    LONG_VERSION: "--keep-output",
    DESCRIPTION: "If set, output sent through the terminal won't substitute lines, looking more like the log."
  }
}



"""
### Mode Compile And Install Executor Module.

This module contains the class to compile and install using CMake.
"""

from typing import Tuple, Dict

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger import predefined_messages, errors
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler
from xmipp3_installer.installer.modes.mode_cmake import mode_cmake_executor

class ModeCompileAndInstallExecutor(mode_cmake_executor.ModeCMakeExecutor):
  """
  ### Mode Compile And Install Executor.

  Compiles and installs using CMake with the appropriate parameters.
  """
  
  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.jobs = context[params.PARAM_JOBS]

  def _set_executor_config(self):
    """### Sets the specific executor params for this mode."""
    super()._set_executor_config()
    self.prints_banner_on_exit = True

  def _run_cmake_mode(self, cmake: str) -> Tuple[int, str]:
    """
    ### Runs the CMake compilation & installation with the appropiate params.

    #### Params:
    - cmake (str): Path to CMake executable.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error. 
    """
    logger(predefined_messages.get_section_message("Compiling with CMake"))
    cmd = f"{cmake} --build {paths.BUILD_PATH} --config {self.build_type} -j {self.jobs}"
    ret_code = shell_handler.run_shell_command_in_streaming(cmd, show_output=True, substitute=self.substitute)
    if ret_code:
      return self._get_error_code(ret_code, errors.CMAKE_COMPILE_ERROR), ""
    logger(predefined_messages.get_done_message(), substitute=self.substitute)
    
    installation_section_message = predefined_messages.get_section_message("Installing with CMake")
    logger(f"\n{installation_section_message}")
    cmd = f"{cmake} --install {paths.BUILD_PATH} --config {self.build_type}"
    ret_code = shell_handler.run_shell_command_in_streaming(cmd, show_output=True, substitute=self.substitute)
    if ret_code:
      return self._get_error_code(ret_code, errors.CMAKE_INSTALL_ERROR), ""
    logger(predefined_messages.get_done_message(), substitute=self.substitute)
    return 0, ""



"""### Contains functions that orquestrate other function executions."""

import multiprocessing
from typing import List, Tuple, Callable, Any

def run_parallel_jobs(
  funcs: List[Callable],
  func_args: List[Tuple[Any, ...]],
  n_jobs: int=multiprocessing.cpu_count()
) -> List:
  """
  ### Runs the given command list in parallel.

  #### Params:
  - funcs (list(callable)): Functions to run.
  - func_args (list(tuple(any, ...))): Arguments for each function.
  - n_jobs (int): Optional. Number of parallel jobs.

  #### Returns:
  - (list): List containing the return of each function.
  """
  with multiprocessing.Pool(n_jobs) as p:
    results = p.starmap(__run_lambda, zip(funcs, func_args))
  return results

def __run_lambda(func: Callable, args: Tuple[Any]) -> Any:
  """
  ### Runs the given function with its args.
  
  #### Params:
  - func (callable): Function to run.
  - args (tuple(any)): Arguments for the function.
  
  #### Returns:
  - (any): Return of the called function.
  """
  return func(*args)



"""
### Mode Clean Executor Module.

This module contains the base class for executors that clean compiled binaries.
"""

from abc import abstractmethod
from typing import Tuple, List

from xmipp3_installer.application import user_interactions
from xmipp3_installer.application.logger import errors
from xmipp3_installer.application.logger import predefined_messages
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.modes import mode_executor
from xmipp3_installer.shared import file_operations

class ModeCleanExecutor(mode_executor.ModeExecutor):
  """
  ### Mode Clean Executor.

  Base class for executors that clean compiled binaries.
  """
  
  def run(self) -> Tuple[int, str]:
    """
    ### Deletes the compiled binaries.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error. 
    """
    if not self.__get_confirmation():
      return errors.INTERRUPTED_ERROR, ""
    file_operations.delete_paths(self.__class__._get_paths_to_delete())
    logger(predefined_messages.get_done_message())
    return 0, ""
  
  def __get_confirmation(self) -> bool:
    """
    ### Asks the user for confirmation.

    #### Returns:
    - (bool): True if the user confirms, False otherwise.
    """
    logger(self._get_confirmation_message())
    return user_interactions.get_user_confirmation(self.confirmation_keyword)
  
  @property
  @abstractmethod
  def confirmation_keyword(self) -> str:
    """### Confirmation keyword propery to be defined by inheriting classes."""

  @staticmethod
  @abstractmethod
  def _get_paths_to_delete() -> List[str]:
    """Get paths to delete method to be implemented by the inheriting classes."""
  
  @abstractmethod
  def _get_confirmation_message(self) -> str:
    """Get confirmation message method to be implemented by the inheriting classes."""



"""
### Mode All Executor Module.

This module contains the class to run the whole installation process.
"""

from typing import Tuple, Dict

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.modes import (
  mode_executor, mode_config_executor, mode_get_sources_executor
)
from xmipp3_installer.installer.modes.mode_cmake import (
  mode_config_build_executor, mode_compile_and_install_executor
)

class ModeAllExecutor(mode_executor.ModeExecutor):
  """
  ### Mode All Executor.

  Runs the whole installation process with the appropriate parameters.
  """
  
  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    config_executor = mode_config_executor.ModeConfigExecutor(
      {**context, params.PARAM_OVERWRITE: False}
    )
    get_sources_executor = mode_get_sources_executor.ModeGetSourcesExecutor(
      context
    )
    config_build_executor = mode_config_build_executor.ModeConfigBuildExecutor(
      context
    )
    compile_and_install_executor = mode_compile_and_install_executor.ModeCompileAndInstallExecutor(
      context
    )
    self.executors = [
      config_executor,
      get_sources_executor,
      config_build_executor,
      compile_and_install_executor
    ]
    super().__init__(context)

  def _set_executor_config(self):
    """### Sets the specific executor params for this mode."""
    super()._set_executor_config()
    self.logs_to_file = True
    self.prints_with_substitution = True
    self.prints_banner_on_exit = True
    self.sends_installation_info = True

  def run(self) -> Tuple[int, str]:
    """
    ### Runs the whole installation process with the appropiate params.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error. 
    """
    for executor_index in range(len(self.executors)):
      if executor_index != 0:
        logger("")
      ret_code, output = self.executors[executor_index].run()
      if ret_code:
        return ret_code, output
    return 0, ""



"""### Contains constants used to configure the installer."""

# Repository names
XMIPP = 'xmipp'
XMIPP_CORE = 'xmippCore'
XMIPP_VIZ = 'xmippViz'
XMIPP_SOURCES = [XMIPP_CORE, XMIPP_VIZ]

# Branch names
MAIN_BRANCHNAME = 'main'

# Context
VERSIONS_CONTEXT_KEY = "versions"

# Others
TAIL_LOG_NCHARS = 300



"""### Contains all the variables related to the configuration."""

SEND_INSTALLATION_STATISTICS = 'SEND_INSTALLATION_STATISTICS'
CMAKE = 'CMAKE'
CC = 'CMAKE_C_COMPILER'
CXX = 'CMAKE_CXX_COMPILER'
CC_FLAGS = 'CMAKE_C_FLAGS'
CXX_FLAGS = 'CMAKE_CXX_FLAGS'
CMAKE_INSTALL_PREFIX = 'CMAKE_INSTALL_PREFIX'
CUDA = 'XMIPP_USE_CUDA'
CUDA_COMPILER = 'CMAKE_CUDA_COMPILER'
MPI = 'XMIPP_USE_MPI'
PREFIX_PATH = 'CMAKE_PREFIX_PATH'
MPI_HOME = 'MPI_HOME'
PYTHON_HOME = 'Python3_ROOT_DIR'
FFTW_HOME = 'FFTW_ROOT'
TIFF_HOME = 'TIFF_ROOT'
HDF5_HOME = 'HDF5_ROOT'
JPEG_HOME = 'JPEG_ROOT'
SQLITE_HOME = 'SQLite_ROOT'
CUDA_CXX = 'CMAKE_CUDA_HOST_COMPILER'
MATLAB = 'XMIPP_USE_MATLAB'
LINK_SCIPION = 'XMIPP_LINK_TO_SCIPION'
BUILD_TESTING = 'BUILD_TESTING'
SKIP_RPATH='CMAKE_SKIP_RPATH'
BUILD_TYPE = "BUILD_TYPE"

# Not stored in ket=value format
LAST_MODIFIED_KEY = "last_modified"

# File structure
TOGGLES = 'toggles'
LOCATIONS = 'locations'
COMPILATION_FLAGS = 'flags'
CONFIG_VARIABLES = {
  TOGGLES: [
    SEND_INSTALLATION_STATISTICS, CUDA, MPI, MATLAB, LINK_SCIPION, BUILD_TESTING, SKIP_RPATH
  ],
  LOCATIONS: [
    CMAKE, CC, CXX, CMAKE_INSTALL_PREFIX, PREFIX_PATH, MPI_HOME,
    CUDA_COMPILER, PYTHON_HOME, FFTW_HOME, TIFF_HOME, 
     HDF5_HOME, JPEG_HOME, SQLITE_HOME, CUDA_CXX
  ],
  COMPILATION_FLAGS: [CC_FLAGS, CXX_FLAGS, BUILD_TYPE]
}

# Do not pass this variables to CMake, only for installer logic
INTERNAL_LOGIC_VARS = [SEND_INSTALLATION_STATISTICS, CMAKE, BUILD_TYPE]



"""### Contains a custom exception for lines in the configuration file with invalid format."""

from xmipp3_installer.application.logger.logger import logger

class InvalidConfigLineError(RuntimeError):
  """
  ### Custom exception for invalid configuration lines.
  
  Raised when a line in the configuration file does not follow the expected format.
  """
  
  @staticmethod
  def generate_error_message(config_file, line_number, line):
    """
    ### Generates an error message for an invalid configuration line.
    
    #### Params:
    - config_file (str): The name of the configuration file.
    - line_number (int): The line number where the error occurred.
    - line (str): The content of the invalid line.
    
    #### Returns:
    - (str): The generated error message.
    """
    return '\n'.join([
      logger.yellow(f"WARNING: There was an error parsing {config_file} file: "),
      logger.red(f'Unable to parse line {line_number}: {line}'),
      logger.yellow(
        "Contents of config file won't be read, default values will be used instead.\n"
        f"You can create a new file template from scratch deleting file {config_file} and trying again."
      )
    ])



"""# Module containing file-related functions."""

import os
import shutil
from typing import List

def delete_paths(paths: List[str]):
  """
  ### Deletes all the given paths (files or directories).

  #### Params:
  - path (list(str)): List of paths to delete.
  """
  for path in paths:
    if not os.path.exists(path):
      continue
    if os.path.isdir(path):
      shutil.rmtree(path, ignore_errors=True)
    else:
      os.remove(path)



"""### URLs."""

DOCUMENTATION_URL = 'https://i2pc.github.io/docs/'
CMAKE_INSTALL_DOCS_URL = 'https://i2pc.github.io/docs/Installation/Requirements/index.html#cmake'
API_URL = 'https://xmipp.i2pc.es/api/attempts/'
MODELS_URL= "http://scipion.cnb.csic.es/downloads/scipion/software/em"
SCIPION_TESTS_URL = "http://scipion.cnb.csic.es/downloads/scipion/data/tests"
I2PC_REPOSITORY_URL = "https://github.com/i2pc/"
DLTK_DOCS_URL = "https://i2pc.github.io/docs/Installation/Optionals/index.html#deeplearningtoolkit"



"""
### Mode Executor Module.

This module contains the base executor interface for installer modes.
"""

from abc import ABC, abstractmethod
from typing import Dict, Tuple

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.constants import paths

class ModeExecutor(ABC):
  """### Base executor interface for installer modes."""

  def __init__(self, context: Dict):
    """
    ### Constructor.
		
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__()
    self.context = context
    self._set_executor_config()
    self.__configure_logger()
  
  def _set_executor_config(self):
    """### Sets the specific executor params for this mode."""
    self.logs_to_file = False
    self.prints_with_substitution = False
    self.prints_banner_on_exit = False
    self.sends_installation_info = False
  
  def __configure_logger(self):
    """### Configures the logger according to the specified config."""
    if self.logs_to_file:
      logger.start_log_file(paths.LOG_FILE)
    if self.prints_with_substitution:
      logger.set_allow_substitution(True)
  
  @abstractmethod
  def run(self) -> Tuple[int, str]:
    """Run method to be implemented by inheriting classes."""



"""### Containis all common constants needed for the argument parsing part of Xmipp's installation."""

from xmipp3_installer.installer.constants import paths

XMIPP_PROGRAM_NAME = "xmipp"

# Other variables
COMMON_USAGE_HELP_MESSAGE = 'Run \"./xmipp -h\" for usage help.'
DEFAULT_MODELS_DIR = paths.INSTALL_PATH



"""
### Mode Get Sources Executor Module.

This module contains the class to clone or update Xmipp source repositories.
"""

import os
from typing import Dict, Tuple, Optional

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger import predefined_messages, errors
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import constants, urls
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes import mode_executor
from xmipp3_installer.installer.handlers import git_handler, versions_manager, shell_handler

class ModeGetSourcesExecutor(mode_executor.ModeExecutor):
  """
  ### Mode Get Sources Executor.

  Clones or updates Xmipp source repositories.
  """

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    - substitute (bool): Optional. If True, printed text will be substituted with the next message.
    """
    super().__init__(context)
    self.substitute = not context[params.PARAM_KEEP_OUTPUT]
    self.target_branch = context.pop(params.PARAM_BRANCH)
    versions: versions_manager.VersionsManager = context[constants.VERSIONS_CONTEXT_KEY]
    self.xmipp_tag_name = versions.xmipp_version_name
    self.source_versions: Dict = versions.sources_versions
  
  def run(self) -> Tuple[int, str]:
    """
    ### Clones or updates Xmipp source repositories.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    logger(predefined_messages.get_section_message("Getting Xmipp sources"))
    for source in constants.XMIPP_SOURCES:
      ret_code, output = self.__get_source(source)
      if ret_code:
        return errors.SOURCE_CLONE_ERROR, output
    return 0, ""
  
  def _set_executor_config(self):
    """### Sets the specific executor params for this mode."""
    super()._set_executor_config()
    self.prints_with_substitution = True
  
  def __select_ref_to_clone(self, source_name: str, source_repo: str) -> Optional[str]:
    """
    ### Selects the reference to clone from the source.

    #### Params:
    - source_name (str): Name of the source to clone.
    - source_repo (str): URL of the source's repository.

    #### Returns:
    - (str | None): The reference name to clone or None if no suitable ref was found.
    """
    current_branch = git_handler.get_current_branch()
    tag_name = None
    if (not current_branch or current_branch == self.xmipp_tag_name):
      tag_name = self.source_versions.get(source_name)
    return git_handler.get_clonable_branch(source_repo, self.target_branch, tag_name)
  
  def __get_source(self, source_name: str) -> Tuple[int, str]:
    """
    ### Gets the given source.
    
    It is cloned if it does not already exist locally.

    #### Params:
    - source_name (str): Name of the source to clone.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and the text output produced by the command.
    """
    repo_url = f"{urls.I2PC_REPOSITORY_URL}{source_name}"
    logger(f"Cloning {source_name}...", substitute=self.substitute)
    logger(predefined_messages.get_working_message(), substitute=self.substitute)

    clone_branch = self.__select_ref_to_clone(source_name, repo_url)
    if self.target_branch and not clone_branch:
      warning_message = "\n".join([
        logger.yellow(f"Warning: branch \'{self.target_branch}\' does not exist for repository with url {repo_url}"),
        logger.yellow("Falling back to repository's default branch.")
      ])
      logger(warning_message, substitute=self.substitute)
    
    ret_code, output = _run_source_command(source_name, repo_url, clone_branch)
    if not ret_code:
      logger(predefined_messages.get_done_message(), substitute=self.substitute)
    return ret_code, output

def _run_source_command(source_name: str, source_repo: str, target_branch: Optional[str]) -> Tuple[int, str]:
  """
  ### Executes git clone/checkout commands for a source repository.
  
  If the source already exists locally:
  - If target_branch is specified, checks out that branch.
  - If no target_branch, returns success without changes.
  
  If the source doesn't exist:
  - Clones the repository with the specified branch.
  - If no branch specified, clones with default branch.

  #### Params:
  - source_name (str): Name of the source repository.
  - source_repo (str): URL of the git repository to clone from.
  - target_branch (str | None): Branch or tag to checkout/clone.

  #### Returns:
  - (tuple(int, str)): Tuple containing the return code and the text output produced by the command.
  """
  source_path = paths.get_source_path(source_name)
  if os.path.exists(source_path):
    if not target_branch:
      return 0, ""
    return shell_handler.run_shell_command(
      f"git checkout {target_branch}",
      cwd=source_path
    )
  
  branch_str = (
    f"{params.PARAMS[params.PARAM_BRANCH][params.LONG_VERSION]} {target_branch}"
    if target_branch else ""
  )
  return shell_handler.run_shell_command(f"git clone{branch_str} {source_repo}.git", cwd=paths.SOURCES_PATH)



"""### Contains functions to assemble the data dictionary required by the API client."""

import getpass
import hashlib
import platform
import re
import os
from typing import Optional, List, Dict

import distro

from xmipp3_installer.installer import constants, orquestrator
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler, git_handler, versions_manager
from xmipp3_installer.installer.handlers.cmake import cmake_constants, cmake_handler

def get_installation_info(version_manager: versions_manager.VersionsManager, ret_code: int=0) -> Dict:
  """
  ### Creates a dictionary with the necessary data for the API POST message.
  
  #### Params:
  - version_manager (VersionsManager): Object containing all the version-related info.
  - ret_code (int): Optional. Return code for the API request.
  
  #### Return:
  - (dict): Dictionary with the required info.
  """
  library_versions = cmake_handler.get_library_versions_from_cmake_file(
    paths.LIBRARY_VERSIONS_FILE
  )
  environment_info = orquestrator.run_parallel_jobs(
    [
      __get_cpu_flags,
      git_handler.get_current_branch,
      git_handler.is_branch_up_to_date,
      __is_installed_by_scipion,
      __get_log_tail
    ],
    [(), (), (), (), ()]
  )

  return {
    "user": {
      "userId": __get_user_id()
    },
    "version": {
      "os": get_os_release_name(),
      "cpuFlags": environment_info[0],
      "cuda": library_versions.get(cmake_constants.CMAKE_CUDA),
      "cmake": library_versions.get(cmake_constants.CMAKE_CMAKE),
      "gcc": library_versions.get(cmake_constants.CMAKE_GCC),
      "gpp": library_versions.get(cmake_constants.CMAKE_GPP),
      "mpi": library_versions.get(cmake_constants.CMAKE_MPI),
      "python": library_versions.get(cmake_constants.CMAKE_PYTHON),
      "sqlite": library_versions.get(cmake_constants.CMAKE_SQLITE),
      "java": library_versions.get(cmake_constants.CMAKE_JAVA),
      "hdf5": library_versions.get(cmake_constants.CMAKE_HDF5),
      "jpeg": library_versions.get(cmake_constants.CMAKE_JPEG)
    },
    "xmipp": {
      "branch": __get_installation_branch_name(
        environment_info[1], version_manager
      ),
      "updated": environment_info[2],
      "installedByScipion": environment_info[3]
    },
    "returnCode": ret_code,
    "logTail": environment_info[4] if ret_code else None # Only needed if something went wrong
  }

def get_os_release_name() -> str:
  """
  ### Returns the name of the current system OS release.

  #### Returns:
  - (str): OS release name.
  """
  platform_system = platform.system()
  if platform_system == "Linux":
    return f"{distro.name()} {distro.version()}"
  return f"{platform_system} {platform.release()}"

def __get_installation_branch_name(branch_name: str, version_manager: versions_manager.VersionsManager) -> str:
  """
  ### Returns the branch or release name of Xmipp.

  #### Params:
  - branch_name (str): Retrieved branch name.
  - version_manager (VersionsManager): Object containing all the version-related info.

  #### Return:
  - (str): Release name if Xmipp is in a release branch, or the branch name otherwise.
  """
  return branch_name if (branch_name and not git_handler.is_tag()) else version_manager.xmipp_version_name

def __get_user_id() -> str:
  """
  ### Returns the unique user id for this machine.
  
  #### Returns:
  - (str): User id, or 'Anoymous' if there were any errors.
  """
  identifier = __get_mac_address()
  if not identifier:
    identifier = getpass.getuser()
    if not identifier:
      return "Anonymous"
  
  sha256 = hashlib.sha256()
  sha256.update(identifier.encode())
  return sha256.hexdigest()

def __get_cpu_flags() -> List[str]:
  """
  ### This function obtains the list of compilation flags supported by the CPU.

  #### Returns:
  - (list(str)): List of flags supported by the CPU.  
  """
  flags_header = "Flags:"
  ret_code, flags_line = shell_handler.run_shell_command(f'lscpu | grep \"{flags_header}\"')
  if ret_code:
    return []
  flags_line = flags_line.replace(flags_header, "").strip()
  return [flag for flag in flags_line.split(" ") if flag]

def __get_log_tail() -> Optional[str]:
  """
  ### Returns the last lines of the installation log.
  
  #### Returns:
  - (str | None): Installation log's last lines, or None if there were any errors.
  """
  ret_code, output = shell_handler.run_shell_command(
    f"tail -n {constants.TAIL_LOG_NCHARS} {paths.LOG_FILE}"
  )
  return output if ret_code == 0 else None

def __get_mac_address() -> Optional[str]:
  """
  ### Returns a physical MAC address for this machine. It prioritizes ethernet over wireless.
  
  #### Returns:
  - (str | None): MAC address, or None if there were any errors.
  """
  ret_code, output = shell_handler.run_shell_command("ip addr")
  return __find_mac_address_in_lines(output.split('\n')) if ret_code == 0 else None

def __find_mac_address_in_lines(lines: List[str]) -> Optional[str]:
  """
  ### Returns a physical MAC address within the text lines provided.

  #### Params:
  - lines (list(str)): Lines of text where MAC address should be looked for.
  
  #### Returns:
  - (str | None): MAC address if found, None otherwise.
  """
  for line_index in range(len(lines) - 1):
    match = re.match(r"^\d+: (enp|wlp|eth|ens|eno)\w+", lines[line_index])
    if not match:
      continue
    mac_match = re.search(r"link/ether ([0-9a-f:]{17})", lines[line_index + 1])
    if mac_match:
      return mac_match.group(1)

def __is_installed_by_scipion() -> bool:
  """
  ### Checks if the current xmipp installation is being carried out from Scipion.

  #### Returns:
  - (bool): True if the installation is executed by Scipion, False otherwise.
  """
  return bool(os.getenv("SCIPION_SOFTWARE"))



"""### Help formatter specific for non-generic usage modes."""

from typing import List, Union, cast

from xmipp3_installer.application.cli import arguments
from xmipp3_installer.application.cli.arguments import modes, params
from xmipp3_installer.application.cli.parsers import format
from xmipp3_installer.application.cli.parsers.base_help_formatter import BaseHelpFormatter
from xmipp3_installer.application.logger.logger import logger

class ModeHelpFormatter(BaseHelpFormatter):
  """### Overrides the default help formatter to display a custom help message deppending on the mode selected."""

  def format_help(self):
    """### This method prints the help message of the argument parser."""
    mode = self.__get_mode()
    help_message = f"{self._get_mode_help(mode, general=False)}\n\n"
    help_message += self.__get_args_message(mode)
    help_message += self.__get_examples_message(mode)
    return format.get_formatting_tabs(help_message)

  def __get_mode(self):
    """
    ### Returns the execution mode.

    #### Returns:
    - (str): Execution mode.
    """
    # Retrieved from the parent help message
    # Message received is the format_help output of the main parser's
    # formatter, adding the mode at the end
    return self._prog.split(' ')[-1]
  
  def __get_args_message(self, mode: str) -> str:
    """
    ### Returns the help section containing all the parameters.

    #### Params:
    - mode (str): Usage mode selected.

    #### Returns:
    - (str): Help section containing all parameters.
    """
    args = modes.MODE_ARGS[mode]
    help_message = ''
    options_str = ''
    separator = ''
    
    if len(args) > 0:
      arg_names = [
        self._get_param_first_name(arg_name) for arg_name in self.__flatten_args(args)
      ]
      if self._has_mutually_exclusive_groups(args):
        help_message += logger.yellow(
          "Important: In this mode, there are mutually exclusive groups of params. "
          "You can only use from one of them at a time.\n"
          )
      if self.__args_contain_optional(arg_names):
        help_message += logger.yellow("Note: only params starting with '-' are optional. The rest are required.\n")
      options_str = ' [options]'
      separator = self._get_help_separator() + '\t# Options #\n\n'

    help_message += f'Usage: {arguments.XMIPP_PROGRAM_NAME} {mode}{options_str}\n{separator}'
    help_message += self.__get_args_info(args)
    return help_message

  @staticmethod
  def __args_contain_optional(arg_names: List[str]) -> bool:
    """
    ### Returns True if the param name list contains at least one optional param.

    ### Params:
    - arg_names (list[str]): List containing the param names.

    ### Returns:
    - (bool): True if there is at least one optional param. False otherwise.
    """
    for name in arg_names:
      if name.startswith('-'):
        return True
    return False

  def __get_args_info(self, args: List[Union[str, List[str]]]) -> str:
    """
    ### Returns the info of each param.

    #### Params:
    - args (list[str]): List of parameters.

    #### Returns:
    - (str): Info of all parameters.
    """
    if not self._has_mutually_exclusive_groups(args):
      return self.__get_args_group_info(cast(List[str], args))
    return "\t---------------\n".join([self.__get_args_group_info(cast(List[str], group)) for group in args])

  def __get_args_group_info(self, args: List[str]) -> str:
    """
    ### Returns the info of each param.

    #### Params:
    - args (list[str]): List of parameters.

    #### Returns:
    - (str): Info of all parameters.
    """
    help_message = ''
    for arg in args:
      help_message += self._text_with_limits(
        '\t' + ', '.join(format.get_param_names(arg)),
        params.PARAMS[arg][params.DESCRIPTION]
      )
    return help_message

  @staticmethod
  def __get_examples_message(mode: str) -> str:
    """
    ### Returns the message section containig usage examples.

    #### Params:
    - mode (str): Usage mode selected.

    #### Returns:
    - (str): Message section containing usage examples.
    """
    help_message = ''
    examples = modes.MODE_EXAMPLES[mode]
    for i in range(len(examples)):
      number_str = '' if len(examples) == 1 else f' {i+1}'	
      help_message += f"\nExample{number_str}: {examples[i]}"
    
    if len(examples) > 0:
      help_message += '\n'

    return help_message

  def __flatten_args(self, args: List[Union[str, List[str]]]) -> List[str]:
    """
    ### Flattens a list of arguments.

    #### Params:
    - args (list[str | list[str]]): A list of arguments, which can include nested lists representing mutually exclusive groups.

    #### Returns:
    - (list[str]): A flattened list of arguments, where nested groups are expanded into a single list.
    """
    if not self._has_mutually_exclusive_groups(args):
      return cast(List[str], args)

    return [
      arg
      for arg_group in args
      for arg in arg_group
    ]



"""
### Predefined Messages Module.

This module contains functions to generate standard log messages.
"""

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.handlers import git_handler

__SECTION_MESSAGE_LEN = 60

def get_done_message() -> str:
  """
  ### Returns the standard message string to use when a task is completed.

  #### Returns:
  - (str): Task completion message.
  """
  return logger.green("Done")

def get_working_message() -> str:
  """
  ### Returns the standard message string to use when a task is in progress.

  #### Returns:
  - (str): Task in progress message.
  """
  return logger.yellow("Working...")

def get_section_message(text: str) -> str:
  """
  ### Returns the given text as a section header.

  #### Params:
  - text (str): Title of the section.

  #### Returns:
  - (str): Formatted section header.
  """
  minimum_remaining_len = len("-  -")
  text_len = len(text)
  remaining_len = __SECTION_MESSAGE_LEN - text_len
  if remaining_len < minimum_remaining_len:
    return text
  
  n_dashes = remaining_len - 2
  n_final_dashes = int(n_dashes / 2)
  n_initial_dashes = n_dashes - n_final_dashes
  final_dashes = ''.join(['-' for _ in range(n_final_dashes)])
  initial_dashes = ''.join(['-' for _ in range(n_initial_dashes)])
  return f"{initial_dashes} {text} {final_dashes}"

def get_success_message(tag_version: str) -> str:
  """
  ### Returns the message shown when Xmipp is compiled successfully.

  #### Params:
  - tag_version (str): Version number of the latest release.
  
  #### Returms:
  - (str): Success message.
  """
  release_name = tag_version if git_handler.is_tag() else git_handler.get_current_branch()

  box_wrapper = '*  *'
  half_len_box_wrapper = int(len(box_wrapper) / 2)
  release_message = f'Xmipp {release_name} has been successfully installed, enjoy it!'
  total_len = len(release_message) + len(box_wrapper)
  release_message = f'{box_wrapper[:half_len_box_wrapper]}{logger.green(release_message)}{box_wrapper[half_len_box_wrapper:]}'
  margin_line = f"*{' ' * (total_len - 2)}*"
  box_border = '*' * total_len

  return '\n'.join(["", box_border, margin_line, release_message, margin_line, box_border])



"""### Defines a base help formatter with extened functions to be used by the custom formatters."""

import argparse
import shutil
from typing import List, Tuple, Union

from xmipp3_installer.application.cli.arguments import modes, params
from xmipp3_installer.application.cli.parsers import format

class BaseHelpFormatter(argparse.HelpFormatter):
  """### Extendes the available functions of the generic help formatter."""

  __SECTION_N_DASH = 68
  __SECTION_SPACE_MODE_HELP = 2
  __SECTION_HELP_START = format.TAB_SIZE + __SECTION_N_DASH + __SECTION_SPACE_MODE_HELP
  __LINE_SIZE_LOWER_LIMIT = int(__SECTION_HELP_START * 1.5)

  @staticmethod
  def _get_mode_help(mode: str, general: bool=True) -> str:
    """
    ### Returns the help message of a given mode.

    ### Params:
    - mode (str): Mode to get help text for.
    - general (bool). Optional. If True, only the general help message is displayed.

    ### Returns:
    - (str): Help of the mode (empty if mode not found).
    """
    for group in modes.MODES.keys():
      if mode in modes.MODES[group].keys():
        messages = modes.MODES[group][mode]
        return BaseHelpFormatter.__get_message_from_list(messages, general)
    return ''

  @staticmethod
  def _get_param_first_name(param_key: str) -> str:
    """
    ### Returns the first name of the given param key. Short name has priority over long name.

    ### Params:
    - param_key (str): Key to identify the param.

    ### Returns:
    - (str): Formatted text.
    """
    param = params.PARAMS[param_key]
    return param.get(params.SHORT_VERSION, param.get(params.LONG_VERSION, ''))

  def _get_help_separator(self) -> str:
    """
    ### Returns the line that separates sections inside the help message.

    ### Returns:
    - (str): Line that separates sections inside the help message.
    """
    dashes = ['-' for _ in range(self.__SECTION_N_DASH)]
    return format.get_formatting_tabs(f"\t{''.join(dashes)}\n")

  def _text_with_limits(self, previous_text: str, text: str) -> str:
    """
    ### Returns the given text, formatted so that it does not exceed the character limit by line for the param help section.

    ### Params:
    - previous_text (str): Text inserted before the one to be returned.
    - text (str): The text to be formatted.

    ### Returns:
    - (str): Formatted text.
    """
    remaining_space, fill_in_space = self.__get_spaces(previous_text)
    formatted_help = BaseHelpFormatter.__multi_line_help_text(
      text,
      remaining_space,
      self.__get_start_section_fill_in_space('')
    )
    return f"{previous_text}{fill_in_space}{formatted_help}\n"

  @staticmethod
  def _has_mutually_exclusive_groups(arg_list: List[Union[str, List[str]]]) -> bool:
    """
    ### Checks if the param list provided contains mutually exclusive groups.

    ### Params:
    - arg_list (list[str | list[str]]): List of arguments, which can be strings or lists of strings.

    ### Returns:
    - (bool): True if all elements in the list are strings, False otherwise.
    """
    return any(isinstance(arg, list) for arg in arg_list)

  @staticmethod
  def __get_text_length(text: str) -> int:
    """
    ### Returns the length of a text that might contain tabs.

    #### Params:
    - text (str): Text to measure.

    #### Returns:
    - (int): Text's length.
    """
    return len(format.get_formatting_tabs(text))

  @staticmethod
  def __get_message_from_list(messages: List[str], only_general: bool) -> str:
    """
    ### Return the appropiate message given a list of them and a condition.

    #### Params:
    - messages (list[str]): List of messages.
    - only_general (bool): If True, only the general (first) message is returned.

    #### Returns:
    - (str): Expected messages in a string.
    """
    return messages[0] if only_general else '\n'.join(messages)

  def __get_line_size(self) -> int:
    """
    ### Returns the maximum size for a line.

    ### Returns:
    - (int): Maximum line size.
    """
    size = shutil.get_terminal_size().columns
    return self.__LINE_SIZE_LOWER_LIMIT if size < self.__LINE_SIZE_LOWER_LIMIT else size

  @staticmethod
  def __multi_line_help_text(text: str, size_limit: int, left_fill: str) -> str:
    """
    ### This function returns the given text, formatted in several lines so that it does not exceed the given character limit.

    ### Params:
    - text (str): The text to be formatted.
    - size_limit (int): Size limit for the text.
    - left_fill (str): String to add at the left of each new line.

    ### Returns:
    - (str): Formatted text.
    """
    return (
      text
      if len(text) <= size_limit else
      BaseHelpFormatter.__format_text_in_lines(text, size_limit, left_fill)
    )

  @staticmethod
  def __fit_words_in_line(words: List[str], size_limit: int) -> Tuple[str, List[str]]:
    """
    ### Returns a tuple containig a line with the words from the given list that could fit given the size limit, and the list with the remaining words.

    ### Params:
    - words (list[str]): List of words to try to fit into a line.
    - size_limit (int): Size limit for the text.

    ### Returns:
    - (str): Line with the words that were able to fit in it.
    - (list[str]): List containing the words that could not fit in the line.
    """
    line = ''
    remaining_words = words
    for word in words:
      if BaseHelpFormatter.__word_fits_in_line(line, word, size_limit):
        line, remaining_words = BaseHelpFormatter.__add_word_to_line(line, word, remaining_words)
      else:
        break
    return line, remaining_words

  @staticmethod
  def __word_fits_in_line(line: str, word: str, size_limit: int) -> bool:
    """
    ### Checks if a word can fit in the current line without exceeding the size limit.

    ### Params:
    - line (str): The current line of text.
    - word (str): The word to check.
    - size_limit (int): The maximum allowed size for the line.

    ### Returns:
    - (bool): True if the word fits in the line, False otherwise.
    """
    if line:
      return len(f"{line} {word}") <= size_limit
    return len(word) <= size_limit

  @staticmethod
  def __add_word_to_line(line: str, word: str, remaining_words: List[str]) -> Tuple[str, List[str]]:
    """
    ### Adds a word to the current line and updates the list of remaining words.

    ### Params:
    - line (str): The current line of text.
    - word (str): The word to add to the line.
    - remaining_words (list[str]): The list of words yet to be added to the line.

    ### Returns:
    - (str): The updated line with the new word added.
    - (list[str]): The updated list of remaining words.
    """
    if line:
      line += f" {word}"
    else:
      line = word
    return line, remaining_words[1:]

  @staticmethod
  def __format_text_in_lines(text: str, size_limit: int, left_fill: str):
    """
    ### Returns the text formatted into size-limited lines.

    #### Params:
    - text (str): Text to format.
    - size_limit (int): Max number of characters allowed in a single line.
    - left_fill (str): Starting characters of each line.

    #### Returns:
    - (str): Text formatted into lines.
    """
    words = text.split(' ')
    lines = []
    while words:
      iteration_size_limit = size_limit if size_limit >= len(words[0]) else len(words[0])
      line, words = BaseHelpFormatter.__fit_words_in_line(words, iteration_size_limit)
      line = left_fill + line if lines else line
      lines.append(line)
    return '\n'.join(lines)

  def __get_spaces(self, start_section_text: str) -> Tuple[int, str]:
    if self.__is_start_section_text_exceeding_size_limit(start_section_text):
      # If text exceeds size limit, it means that section space for modes and params 
      # is too low and should be set to a higher number, but for now we need to print anyways, 
      # so we reduce space from the one reserved for mode help and add minimum fill-in space
      remaining_space = self.__get_line_size() - BaseHelpFormatter.__get_text_length(start_section_text)
      fill_in_space = ' '
    else:
      remaining_space = self.__get_line_size() - self.__SECTION_HELP_START
      fill_in_space = self.__get_start_section_fill_in_space(start_section_text)
    return remaining_space, fill_in_space

  def __get_start_section_fill_in_space(self, text: str) -> str:
    """
    ### Returns the fill-in space for the start section.

    #### Params:
    - text (str): Text inside the start section.

    #### Returns:
    - (str): The required number of spaces to generate the start section's fill-in.
    """
    return ''.join(
      [' ' for _ in range(self.__SECTION_HELP_START - BaseHelpFormatter.__get_text_length(text))]
    )
  
  def __is_start_section_text_exceeding_size_limit(self, start_section_text: str) -> bool:
    """
    ### Indicates if the given start section text exceedes allowed size limit.

    #### Params:
    - start_section_text (str): Text to measure.

    #### Returns:
    - (bool): True if the text exceedes allowed size limit.
    """
    return BaseHelpFormatter.__get_text_length(start_section_text) >= self.__SECTION_HELP_START



"""
### Mode Selector Module.

This module maps command-line modes to their corresponding executor classes.
"""

from xmipp3_installer.application.cli.arguments import modes
from xmipp3_installer.installer.modes import (
  mode_config_executor, mode_version_executor, mode_git_executor,
  mode_get_sources_executor, mode_all_executor
)
from xmipp3_installer.installer.modes.mode_clean import mode_clean_all_executor, mode_clean_bin_executor
from xmipp3_installer.installer.modes.mode_sync import (
  mode_add_model_executor, mode_get_models_executor, mode_test_executor
)
from xmipp3_installer.installer.modes.mode_cmake import mode_config_build_executor, mode_compile_and_install_executor

MODE_EXECUTORS = {
  modes.MODE_ADD_MODEL: mode_add_model_executor.ModeAddModelExecutor,
  modes.MODE_ALL: mode_all_executor.ModeAllExecutor,
  modes.MODE_CLEAN_ALL: mode_clean_all_executor.ModeCleanAllExecutor,
  modes.MODE_CLEAN_BIN: mode_clean_bin_executor.ModeCleanBinExecutor,
  modes.MODE_COMPILE_AND_INSTALL: mode_compile_and_install_executor.ModeCompileAndInstallExecutor,
  modes.MODE_CONFIG_BUILD: mode_config_build_executor.ModeConfigBuildExecutor,
  modes.MODE_CONFIG: mode_config_executor.ModeConfigExecutor,
  modes.MODE_GET_MODELS: mode_get_models_executor.ModeGetModelsExecutor,
  modes.MODE_GET_SOURCES: mode_get_sources_executor.ModeGetSourcesExecutor,
  modes.MODE_GIT: mode_git_executor.ModeGitExecutor,
  modes.MODE_TEST: mode_test_executor.ModeTestExecutor,
  modes.MODE_VERSION: mode_version_executor.ModeVersionExecutor
}



"""### Command Line Interface that interacts with the installer."""

import argparse
import multiprocessing
import os
import sys
from typing import Dict, Any

from xmipp3_installer.application.cli import arguments
from xmipp3_installer.application.cli.arguments import modes, params
from xmipp3_installer.application.cli.parsers import format
from xmipp3_installer.application.cli.parsers.error_handler_parser import ErrorHandlerArgumentParser
from xmipp3_installer.application.cli.parsers.general_help_formatter import GeneralHelpFormatter
from xmipp3_installer.application.cli.parsers.mode_help_formatter import ModeHelpFormatter
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import installer_service

def main():
  """### Main entry point function that starts the execution."""
  parser = __generate_parser()
  parser = __add_params(parser)
  __add_default_usage_mode()
  args = vars(parser.parse_args())
  __validate_args(args, parser)
  installation_manager = installer_service.InstallationManager(args)
  ret_code = installation_manager.run_installer()
  sys.exit(ret_code)

def __generate_parser() -> argparse.ArgumentParser:
  """
  ### Generates an argument parser for the installer.

  #### Returns:
  - (ArgumentParser): Argument parser.
  """
  return ErrorHandlerArgumentParser(
    prog=arguments.XMIPP_PROGRAM_NAME,
    formatter_class=GeneralHelpFormatter,
  )

def __add_params(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
  """
  ### Inserts the params into the given parser.

  #### Params:
  - parser (ArgumentParser): Argument parser.

  #### Returns:
  - (ArgumentParser): Argument parser with inserted params.
  """
  subparsers = parser.add_subparsers(dest=modes.MODE)
  default_jobs = __get_default_job_number()

  add_model_subparser = subparsers.add_parser(modes.MODE_ADD_MODEL, formatter_class=ModeHelpFormatter)
  __add_params_mode_add_model(add_model_subparser)

  all_subparser = subparsers.add_parser(modes.MODE_ALL, formatter_class=ModeHelpFormatter)
  __add_params_mode_all(all_subparser, default_jobs)

  subparsers.add_parser(modes.MODE_CLEAN_ALL, formatter_class=ModeHelpFormatter)

  subparsers.add_parser(modes.MODE_CLEAN_BIN, formatter_class=ModeHelpFormatter)

  compile_and_install_subparser = subparsers.add_parser(modes.MODE_COMPILE_AND_INSTALL, formatter_class=ModeHelpFormatter)
  __add_params_mode_compile_and_install(compile_and_install_subparser, default_jobs)

  build_config_subparser = subparsers.add_parser(modes.MODE_CONFIG_BUILD, formatter_class=ModeHelpFormatter)
  __add_params_mode_config_build(build_config_subparser)

  config_subparser = subparsers.add_parser(modes.MODE_CONFIG, formatter_class=ModeHelpFormatter)
  __add_params_mode_config(config_subparser)

  get_models_subparser = subparsers.add_parser(modes.MODE_GET_MODELS, formatter_class=ModeHelpFormatter)
  __add_params_mode_get_models(get_models_subparser)

  get_sources_subparser = subparsers.add_parser(modes.MODE_GET_SOURCES, formatter_class=ModeHelpFormatter)
  __add_params_mode_get_sources(get_sources_subparser)

  git_subparser = subparsers.add_parser(modes.MODE_GIT, formatter_class=ModeHelpFormatter)
  __add_params_mode_git(git_subparser)

  test_subparser = subparsers.add_parser(modes.MODE_TEST, formatter_class=ModeHelpFormatter)
  __add_params_mode_test(test_subparser)

  version_subparser = subparsers.add_parser(modes.MODE_VERSION, formatter_class=ModeHelpFormatter)
  __add_params_mode_version(version_subparser)

  return parser

def __add_params_mode_add_model(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "addModel".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_LOGIN))
  subparser.add_argument(*format.get_param_names(params.PARAM_MODEL_PATH))
  subparser.add_argument(*format.get_param_names(params.PARAM_UPDATE), action='store_true')

def __add_params_mode_all(subparser: argparse.ArgumentParser, default_jobs: int):
  """
  ### Adds params for mode "all".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  - default_jobs (int): Default number of jobs to run the task.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_JOBS), type=int, default=default_jobs)
  subparser.add_argument(*format.get_param_names(params.PARAM_BRANCH))
  subparser.add_argument(*format.get_param_names(params.PARAM_KEEP_OUTPUT), action='store_true')

def __add_params_mode_compile_and_install(subparser: argparse.ArgumentParser, default_jobs: int):
  """
  ### Adds params for mode "compileAndInstall".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  - default_jobs (int): Default number of jobs to run the task.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_JOBS), type=int, default=default_jobs)
  subparser.add_argument(*format.get_param_names(params.PARAM_BRANCH))
  subparser.add_argument(*format.get_param_names(params.PARAM_KEEP_OUTPUT), action='store_true')

def __add_params_mode_config_build(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "configBuild".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_KEEP_OUTPUT), action='store_true')

def __add_params_mode_config(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "config".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_OVERWRITE), action='store_true')

def __add_params_mode_get_models(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "getModels".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(
    *format.get_param_names(params.PARAM_MODELS_DIRECTORY),
    default=os.path.abspath(arguments.DEFAULT_MODELS_DIR)
  )

def __add_params_mode_get_sources(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "getSources".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_BRANCH))
  subparser.add_argument(*format.get_param_names(params.PARAM_KEEP_OUTPUT), action='store_true')

def __add_params_mode_git(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "git".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_GIT_COMMAND), nargs='+')

def __add_params_mode_test(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "test".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  group = subparser.add_mutually_exclusive_group(required=True)
  group.add_argument(*format.get_param_names(params.PARAM_TEST_NAMES), nargs='*', default=[])
  group.add_argument(*format.get_param_names(params.PARAM_SHOW_TESTS), action='store_true')
  group.add_argument(*format.get_param_names(params.PARAM_ALL_FUNCTIONS), action='store_true')
  group.add_argument(*format.get_param_names(params.PARAM_ALL_PROGRAMS), action='store_true')

def __add_params_mode_version(subparser: argparse.ArgumentParser):
  """
  ### Adds params for mode "version".

  #### Params:
  - subparser (ArgumentParser): Subparser to add the params to.
  """
  subparser.add_argument(*format.get_param_names(params.PARAM_SHORT), action='store_true')

def __get_default_job_number() -> int:
  """
  ### Gets the default number of jobs to be used by parallelizable tasks.

  Returned number will be 120% of CPU cores, due to not all jobs taking 
  100% of CPU time continuously.

  #### Returns:
  - (int): Default number of jobs.
  """
  return multiprocessing.cpu_count() + int(multiprocessing.cpu_count() * 0.2)

def __add_default_usage_mode():
  """### Sets the usage mode as the default one when a mode has not been specifically provided."""
  no_args_provided = len(sys.argv) == 1
  args_provided = len(sys.argv) > 1
  if no_args_provided or (
    args_provided and __is_first_arg_optional() and not __help_requested()
    ): 
    sys.argv.insert(1, modes.MODE_ALL)

def __is_first_arg_optional() -> bool:
  """
  ### Returns True if the first argument provided is optional.

  #### Returns:
  - (bool): True if the first argument received is optional.
  """
  return sys.argv[1].startswith('-')

def __help_requested() -> bool:
  """
  ### Returns True if help is at least one of the args.

  #### Returns:
  - (bool): True if help is at least one of the args.
  """
  return '-h' in sys.argv or '--help' in sys.argv

def __validate_args(args: Dict[str, Any], parser: argparse.ArgumentParser):
  """
  ### Performs validations on the arguments.

  #### Params:
  - args (dict(str, any)): Arguments to be validated.
  - parser (ArgumentParser): Argument parser.
  """
  jobs = args.get('jobs', 1)
  if jobs < 1:
    parser.error(f"Wrong job number \"{jobs}\". Number of jobs has to be 1 or greater.")
  
  branch = args.get('branch')
  if branch is not None and len(branch.split(' ')) > 1:
    parser.error(f"Incorrect branch name \"{branch}\". Branch names can only be one word long.")
  
  if args.get('keep_output', False):
    logger.set_allow_substitution(False)



"""
### Installer Service Module.

This module contains the class to manage the installation process.
"""

from typing import Dict, cast

from xmipp3_installer.api_client import api_client
from xmipp3_installer.api_client.assembler import installation_info_assembler
from xmipp3_installer.application.cli.arguments import modes
from xmipp3_installer.application.logger import errors
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import constants
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import versions_manager
from xmipp3_installer.installer.modes.mode_executor import ModeExecutor
from xmipp3_installer.application.logger import predefined_messages
from xmipp3_installer.installer.modes import mode_selector
from xmipp3_installer.repository import config
from xmipp3_installer.repository.config_vars import variables

class InstallationManager:
  """
  ### Installation Manager.

  Manages the installation process by executing the appropriate mode based on the given arguments.
  """
  
  def __init__(self, args: Dict):
    """
    ### Constructor.
		
    #### Params:
    - args (dict): Dictionary containing all parsed command-line arguments.
    """
    self.mode = args.pop(modes.MODE, modes.MODE_ALL)
    config_handler = config.ConfigurationFileHandler(path=paths.CONFIG_FILE, show_errors=False)
    self.context = {
      **args,
      **config_handler.values,
      variables.LAST_MODIFIED_KEY: config_handler.last_modified,
      constants.VERSIONS_CONTEXT_KEY: versions_manager.VersionsManager(paths.VERSION_INFO_FILE)
    }
    self.mode_executor: ModeExecutor = mode_selector.MODE_EXECUTORS[self.mode](self.context)

  def run_installer(self):
    """
    ### Runs the installer with the given arguments.

    #### Returns:
    - (int): Return code.
    """
    try:
      ret_code, output = self.mode_executor.run()
    except KeyboardInterrupt:
      logger.log_error("", ret_code=errors.INTERRUPTED_ERROR, add_portal_link=False)
      return errors.INTERRUPTED_ERROR
    if ret_code:
      logger.log_error(output, ret_code=ret_code, add_portal_link=ret_code != errors.INTERRUPTED_ERROR)
    if (
      self.mode_executor.sends_installation_info and 
      self.context[variables.SEND_INSTALLATION_STATISTICS]
    ):
      logger("Sending anonymous installation info...", show_in_terminal=False)
      api_client.send_installation_attempt(
        installation_info_assembler.get_installation_info(
          self.context[constants.VERSIONS_CONTEXT_KEY],
          ret_code=ret_code
        )
      )
    if not ret_code and self.mode_executor.prints_banner_on_exit:
      logger(predefined_messages.get_success_message(
        cast(
          versions_manager.VersionsManager,
          self.context[constants.VERSIONS_CONTEXT_KEY]
        ).xmipp_version_name
      ))
    logger.close()
    return ret_code



"""
### Versions Manager Module.

This module contains the class to manage version information.
"""

import json
from datetime import datetime
from typing import Dict, List

from xmipp3_installer.installer import constants
from xmipp3_installer.shared import singleton

class VersionsManager(singleton.Singleton):
  """
  ### Versions Manager.

  Manages version information by reading and validating version data from a JSON file.
  """

  def __init__(self, version_file_path: str):
    """
    ### Constructor.
    
    Initializes the version manager by reading and validating version information from a JSON file.

    #### Params:
    - version_file_path (str): Path to the JSON file containing version information.

    #### Raises:
    - ValueError: If version numbers don't follow semver format (x.y.z) or release date isn't dd/mm/yyyy.
    """
    super().__init__()
    self.version_file_path = version_file_path
    version_info = self.__get_version_info()
    self.xmipp_version_number = version_info[constants.XMIPP]["version_number"]
    self.xmipp_version_name = version_info[constants.XMIPP]["version_name"]
    self.xmipp_release_date = version_info[constants.XMIPP]["release_date"]
    self.sources_versions = version_info["sources_target_tag"]
    self.__validate_fields()

  def __get_version_info(self) -> Dict[str, Dict[str, str]]:
    """
    ### Retrieves the version info from the version information JSON file.

    #### Returns:
    - (dict(str, dict(str, str))): Dictionary containing the parsed values.
    """
    with open(self.version_file_path, encoding="utf-8") as json_data:
      version_info = json.load(json_data)
    return version_info
  
  def __validate_fields(self):
    """
    ### Validates version numbers and release date format.
    
    Checks that:
    - Version numbers follow semantic versioning (x.y.z)
    - Release date follows dd/mm/yyyy format

    #### Raises:
    - ValueError: If any field doesn't match its required format.
    """
    self.__validate_version_number()
    self.__validate_release_date()

  def __validate_version_number(self):
    """
    ### Validates that version numbers follow semantic versioning format.
    
    Checks that version numbers are in x.y.z format where x, y, z are integers.

    #### Raises:
    - ValueError: If any version number doesn't follow the required format.
    """
    parts = self.xmipp_version_number.split('.')
    if not VersionsManager.__is_valid_semver(parts):
      raise ValueError(
        f"Version number '{self.xmipp_version_number}' is invalid. Must be three numbers separated by dots (x.y.z)."
      )

  @staticmethod
  def __is_valid_semver(version_parts: List[str]) -> bool:
    """
    ### Checks if version parts constitute a valid semantic version.
    
    #### Params:
    - version_parts (list(str)): List of strings representing version number parts.

    #### Returns:
    - (bool): True if version follows semver format, False otherwise.
    """
    SEMVER_N_NUMBERS = 3 # 3 numbers separated by dots: X.Y.Z
    return len(version_parts) == SEMVER_N_NUMBERS and all(part.isdigit() for part in version_parts)

  def __validate_release_date(self):
    """
    ### Validates that release date follows dd/mm/yyyy format.
    
    Checks that:
    - Date string follows dd/mm/yyyy format
    - Date is a valid calendar date

    #### Raises:
    - ValueError: If release date doesn't follow the required format or is invalid.
    """
    try:
      datetime.strptime(self.xmipp_release_date, "%d/%m/%Y")
    except ValueError:
      raise ValueError(
        f"Release date '{self.xmipp_release_date}' is invalid. Must be in dd/mm/yyyy format."
      )



"""
### Paths Module.

This module defines various paths used throughout the application.
"""

import os

from xmipp3_installer.installer import constants

# General paths
SOURCES_PATH = "src"
BUILD_PATH = "build"
INSTALL_PATH = "dist"
BINARIES_PATH = os.path.join(INSTALL_PATH, "bin")
SCIPION_SOFTWARE_EM = os.path.join("scipionfiles", "downloads", "scipion", "software", "em")

# General file paths
LOG_FILE = 'compilation.log'
LIBRARY_VERSIONS_FILE = os.path.join(BUILD_PATH, 'versions.txt')
CONFIG_FILE = 'xmipp.conf'
VERSION_INFO_FILE = "version-info.json"

# Source paths
def get_source_path(source: str) -> str:
  """
  ### Returns the path to the given source.

  #### Params:
  - source (str): Source name.

  #### Returns:
  - (str): Path to the source.
  """
  return os.path.abspath(os.path.join(SOURCES_PATH, source))
XMIPP_CORE_PATH = get_source_path(constants.XMIPP_CORE)
XMIPP_SOURCE_PATHS = [get_source_path(source) for source in constants.XMIPP_SOURCES]



"""### Provides a global logger."""

import shutil
from io import BufferedReader

from xmipp3_installer.shared.singleton import Singleton
from xmipp3_installer.application.logger import errors
from xmipp3_installer.installer import urls

class Logger(Singleton):
  """### Logger class for keeping track of installation messages."""

  __UP = "\x1B[1A\r"
  __REMOVE_LINE = '\033[K'
  __BOLD = "\033[1m"
  __BLUE = "\033[34m"
  __RED = "\033[91m"
  __GREEN = "\033[92m"
  __YELLOW = "\033[93m"
  __END_FORMAT = "\033[0m"
  __FORMATTING_CHARACTERS = [__UP, __REMOVE_LINE, __BOLD, __BLUE, __RED, __GREEN, __YELLOW, __END_FORMAT]
 
  def __init__(self):
    """
    ### Constructor.
    
    #### Params:
    - ouputToConsoloe (bool): Print messages to console.
    """
    self.__log_file = None
    self.__last_printed_elem = None
    self.__allow_substitution = True
  
  def green(self, text: str) -> str:
    """
    ### This function returns the given text formatted in green color.

    #### Params:
    - text (str): Text to format.

    #### Returns:
    - (str): Text formatted in green color.
    """
    return self.__format_text(text, self.__GREEN)

  def yellow(self, text: str) -> str:
    """
    ### This function returns the given text formatted in yellow color.

    #### Params:
    - text (str): Text to format.

    #### Returns:
    - (str): Text formatted in yellow color.
    """
    return self.__format_text(text, self.__YELLOW)

  def red(self, text: str) -> str:
    """
    ### This function returns the given text formatted in red color.

    #### Params:
    - text (str): Text to format.

    #### Returns:
    - (str): Text formatted in red color.
    """
    return self.__format_text(text, self.__RED)

  def blue(self, text: str) -> str:
    """
    ### This function returns the given text formatted in blue color.

    #### Params:
    - text (str): Text to format.

    #### Returns:
    - (str): Text formatted in blue color.
    """
    return self.__format_text(text, self.__BLUE)

  def bold(self, text: str) -> str:
    """
    ### This function returns the given text formatted in bold.

    #### Params:
    - text (str): Text to format.

    #### Returns:
    - (str): Text formatted in bold.
    """
    return self.__format_text(text, self.__BOLD)

  def start_log_file(self, log_path: str):
    """
    ### Initiates the log file.

    #### Params:
    - log_path (str): Path to the log file.
    """
    if self.__log_file is None:
      self.__log_file = open(log_path, 'w', encoding="utf-8")

  def close(self):
    """### Closes the log file."""
    if self.__log_file:
      self.__log_file.close()
      self.__log_file = None

  def set_allow_substitution(self, allow_substitution: bool):
    """
    ### Modifies console output behaviour, allowing or disallowing substitutions.
    
    #### Params:
    - allow_substitution (bool): If False, console outputs won't be substituted.
    """
    self.__allow_substitution = allow_substitution

  def __call__(self, text: str, show_in_terminal: bool=True, substitute: bool=False):
    """
    ### Log a message.
    
    #### Params:
    - text (str): Message to be logged. Supports fancy formatting.
    - show_in_terminal (bool): Optional. If True, text is also printed through terminal.
    - substitute (bool): Optional. If True, previous line is substituted with new text. Only used when show_in_terminal = True.
    """
    if self.__log_file is not None:
      print(self.__remove_non_printable(text), file=self.__log_file, flush=True)
      
    if show_in_terminal:
      text = self.__substitute_lines(text) if self.__allow_substitution and substitute else text
      print(text, flush=True)
    
    # Store printed string for next substitution calculation
    self.__last_printed_elem = self.__remove_non_printable(text) if self.__allow_substitution and substitute else None
   
  def log_error(self, error_msg: str, ret_code: int=1, add_portal_link: bool=True):
    """
    ### Prints an error message.

    #### Params:
    - error_msg (str): Error message to show.
    - ret_code (int): Optional. Return code to end the exection with.
    - add_portal_link (bool): If True, a message linking the documentation portal is shown.
    """
    error = errors.ERROR_CODES.get(ret_code, errors.ERROR_CODES[errors.UNKOW_ERROR])
    error_str = error_msg + '\n\n' if error_msg else ''
    error_str += f'Error {ret_code}: {error[0]}'
    error_str += f"\n{error[1]}" if error[1] else ''
    if add_portal_link:
      error_str += f'\nMore details on the Xmipp documentation portal: {urls.DOCUMENTATION_URL}'

    self.__call__(self.red(error_str), show_in_terminal=True)

  def log_in_streaming(self, stream: BufferedReader, show_in_terminal: bool=False, substitute: bool=False, err: bool=False):
    """
    ### Receives a process output stream and logs its lines.

    #### Params:
    - stream (BufferedReader): Function to run.
    - show_in_terminal (bool): Optional. If True, output will also be printed through terminal.
    - substitute (bool): Optional. If True, output will replace previous line. Only used when show is True.
    - err (bool): Optional. If True, the stream contains an error. Otherwise, it is regular output.
    """
    for line in iter(stream.readline, b''):
      calling_line = line.decode().replace("\n", "")
      if err:
        calling_line = self.red(calling_line)
      self.__call__(calling_line, show_in_terminal=show_in_terminal, substitute=substitute)

  def __remove_non_printable(self, text: str) -> str:
    """
    ### This function returns the given text without non printable characters.

    #### Params:
    - text (str): Text to remove format.

    #### Returns:
    - (str): Text without format.
    """
    for formatting_char in self.__FORMATTING_CHARACTERS:
      text = text.replace(formatting_char, "")
    return text

  def __get_n_last_lines(self) -> int:
    """
    ### This function returns the number of lines of the terminal the last print occupied.

    #### Returns:
    - (int): Number of lines of the last print. 
    """
    if self.__last_printed_elem is None:
      return 0
    
    terminal_width = shutil.get_terminal_size().columns
    # At least one line break exists, as prints end with line break
    n_line_breaks = self.__last_printed_elem.count("\n") + 1
    text_lines = self.__last_printed_elem.split("\n")
    for line in text_lines:
      n_line_breaks += int(len(line) / (terminal_width + 1)) # Equal does not count, it needs to exceed
    return n_line_breaks
  
  def __format_text(self, text: str, format_code: str) -> str:
    """
    ### Returns the given text formatted in the given format code.

    #### Params:
    - text (str): Text to format.
    - format_code (str): Formatting character.

    #### Returns:
    - (str): Text formatted.
    """
    return f"{format_code}{text}{self.__END_FORMAT}"

  def __substitute_lines(self, text: str) -> str:
    """
    ### Removes the appropiate lines from the terminal to show a given text.

    #### Params:
    - text (str): Text to show substituting some lines.

    #### Returns:
    - (str): Text with line substitution characters as a prefix.
    """
    substitution_chars = [f'{self.__UP}{self.__REMOVE_LINE}' for _ in range(self.__get_n_last_lines())]
    return f"{''.join(substitution_chars)}{text}"
  
"""
### Global logger.
"""
logger = Logger()



"""
### Mode Git Executor Module.

This module contains the class to execute git commands on all Xmipp source repositories.
"""

from typing import Dict, Tuple

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import constants
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes import mode_executor
from xmipp3_installer.installer.handlers import git_handler

class ModeGitExecutor(mode_executor.ModeExecutor):
  """
  ### Mode Git Executor.

  Executes git commands on all Xmipp source repositories.
  """

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    command_param_list = context.pop(params.PARAM_GIT_COMMAND)
    self.command = ' '.join(command_param_list)
  
  def run(self) -> Tuple[int, str]:
    """
    ### Executes the given git command into all xmipp source repositories.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    logger(f"Running command 'git {self.command}' for all xmipp sources...")

    for source in [constants.XMIPP, *constants.XMIPP_SOURCES]:
      logger("\n" + logger.blue(
        f"Running command for {source} in path {paths.get_source_path(source)}..."
      ))
      ret_code, output = git_handler.execute_git_command_for_source(
        self.command, source
      )
      if ret_code:
        return ret_code, output

    return 0, ""



"""### Argparser that shows the error messages formatted in a custom way."""

import argparse
from typing import List

from xmipp3_installer.application.cli.parsers import format
from xmipp3_installer.application.logger.logger import logger

class ErrorHandlerArgumentParser(argparse.ArgumentParser):
  """### Overrides the error function of the standard argument parser to display better error messages."""

  def error(self, message):
    """
    ### Prints through stderr the error message and exits with specific return code.
    
    #### Params:
    - message (str): Error message.
    """
    args = self.__get_args()
    mode = ErrorHandlerArgumentParser.__get_mode(args)

    if ErrorHandlerArgumentParser.__is_mode_generic(args):
      args = ' '.join(args[:-1])
      extra_line_break = '\n'
    else:
      args = self.format_help()
      extra_line_break = ''

    error_message = logger.red(f"{mode}: error: {message}\n")
    self.exit(
      1,
      format.get_formatting_tabs(f"{args}{extra_line_break}{error_message}")
    )
  
  def __get_args(self) -> List[str]:
    """
    ### Obtains args from stored class data.

    #### Returns:
    - (list[str]): List of arguments.
    """
    return self.prog.split(' ')
  
  @staticmethod
  def __get_mode(args: List[str]) -> str:
    """
    ### Obtains the usage mode from the received args.

    #### Params:
    - args (list[str]): List of args received by the parser.
    
    #### Returns:
    - (str): Usage mode.
    """
    return args[-1]
  
  @staticmethod
  def __is_mode_generic(args: List[str]) -> bool:
    """
    ### Returns True if the usage mode selected is the generic one.

    #### Params:
    - args (list[str]): List of received args.

    #### Returns:
    - (bool): True if the received mode is the generic one.
    """
    return len(args) > 1



"""### Containis all mode constants needed for the argument parsing part of Xmipp's installation."""

from xmipp3_installer.application.cli import arguments
from xmipp3_installer.application.cli.arguments.params import (
  PARAM_SHORT, PARAM_JOBS, PARAM_BRANCH, PARAM_KEEP_OUTPUT, PARAM_OVERWRITE,
  PARAM_MODELS_DIRECTORY, PARAM_TEST_NAMES, PARAM_SHOW_TESTS, PARAM_GIT_COMMAND,
  PARAM_LOGIN, PARAM_MODEL_PATH, PARAM_UPDATE, PARAMS, PARAM_ALL_FUNCTIONS,
  PARAM_ALL_PROGRAMS, SHORT_VERSION, LONG_VERSION
)
from xmipp3_installer.installer import urls, constants

MODE = "mode"

# Mode list (alphabetical order)
MODE_ADD_MODEL = 'addModel'
MODE_ALL = 'all'
MODE_CLEAN_ALL = 'cleanAll'
MODE_CLEAN_BIN = 'cleanBin'
MODE_COMPILE_AND_INSTALL = 'compileAndInstall'
MODE_CONFIG_BUILD = 'configBuild'
MODE_CONFIG = 'config'
MODE_GET_MODELS = 'getModels'
MODE_GET_SOURCES = 'getSources'
MODE_GIT = 'git'
MODE_TEST = 'test'
MODE_VERSION = 'version'

# Modes with help message
MODES = {
  'General': {
    MODE_VERSION: ['Returns the version information. Add \'--short\' to print only the version number.'],
    MODE_COMPILE_AND_INSTALL: ['Compiles and installs Xmipp based on already obtained sources.'],
    MODE_ALL: [f'Default param. Runs {MODE_CONFIG}, {MODE_CONFIG_BUILD}, and {MODE_COMPILE_AND_INSTALL}.'],
    MODE_CONFIG_BUILD: ['Configures the project with CMake.']
  },
  'Config': {
    MODE_CONFIG: ['Generates a config file template with default values.'],
  },
  'Downloads': {
    MODE_GET_MODELS: [
      f'Downloads the Deep Learning Models required by the DLTK tools at dir/models ({arguments.DEFAULT_MODELS_DIR} by default).',
      f'Visit {urls.DLTK_DOCS_URL} for more details.'
    ],
    MODE_GET_SOURCES: [f'Clones Xmipp\'s source repositories {constants.XMIPP_CORE} & {constants.XMIPP_VIZ}.']
  },
  'Clean': {
    MODE_CLEAN_BIN: ['Removes all compiled binaries.'],
    MODE_CLEAN_ALL: ['Removes all compiled binaries and sources, leaves the repository as if freshly cloned (without pulling).']
  },
  'Test': {
    MODE_TEST: [
      'Runs Xmipp\'s tests.',
      f'If used with \'{PARAMS[PARAM_TEST_NAMES][SHORT_VERSION]}\', only the tests provided will run.',
      f'If used with \'{PARAMS[PARAM_SHOW_TESTS][LONG_VERSION]}\', a list the tests available and how to invoke them will be shown.',
      f'If used with \'{PARAMS[PARAM_ALL_FUNCTIONS][LONG_VERSION]}\', all function tests will run.',
      f'If used with \'{PARAMS[PARAM_ALL_PROGRAMS][LONG_VERSION]}\', all program tests will run.'
    ]
  },
  'Developers': {
    MODE_GIT: ['Runs the given git action for all source repositories.'],
    MODE_ADD_MODEL: [
      "Takes a DeepLearning model from the modelPath, makes a tgz of it and uploads the .tgz according to the <login>.",
      "This mode is used to upload a model folder to the Scipion/Xmipp server.",
      "Usually the model folder contains big files used to feed deep learning procedures "
      "with pretrained data. All the models stored in the server will be downloaded "
      "using the 'get_models' mode, during the compilation/installation time, "
      "or with scipion3 installb deepLearningToolkit.",
      "Param <modelsPath> must be an absolute path.",
      "",
      "Usage: -> ./xmipp addModel <usr@server> <modelsPath> [--update]",
      "Steps:	0. modelName = basename(modelsPath) <- Please, check the folders name!",
      "        1. Packing in 'xmipp_model_modelName.tgz'",
      "        2. Check if that model already exists (use --update to override an existing model)",
      "        3. Upload the model to the server.",
      "        4. Update the MANIFEST file.",
      "",
      "The model name will be the folder name in <modelsPath>",
      "Must have write permissions to such machine."
    ]
  }
}

# Arguments of each mode, sorted by group
# Inside each mode, params are grouped by mutually exclusive groups
# For example: MYMODE: [[PARAM1, PARAM2], [PARAM3, PARAM4]] would translate to
# "mymode ([param1] [param2] | [param3] [param4])" in the general help message
MODE_ARGS = {
  MODE_VERSION: [PARAM_SHORT],
  MODE_COMPILE_AND_INSTALL: [PARAM_JOBS, PARAM_KEEP_OUTPUT],
  MODE_ALL: [PARAM_JOBS, PARAM_BRANCH, PARAM_KEEP_OUTPUT],
  MODE_CONFIG_BUILD: [PARAM_KEEP_OUTPUT],
  MODE_CONFIG: [PARAM_OVERWRITE],
  MODE_GET_MODELS: [PARAM_MODELS_DIRECTORY],
  MODE_GET_SOURCES: [PARAM_BRANCH, PARAM_KEEP_OUTPUT],
  MODE_CLEAN_BIN: [],
  MODE_CLEAN_ALL: [],
  MODE_TEST: [[PARAM_TEST_NAMES], [PARAM_SHOW_TESTS], [PARAM_ALL_FUNCTIONS], [PARAM_ALL_PROGRAMS]],
  MODE_GIT: [PARAM_GIT_COMMAND],
  MODE_ADD_MODEL: [PARAM_LOGIN, PARAM_MODEL_PATH, PARAM_UPDATE]
}

# Examples for the help message of each mode
MODE_EXAMPLES = {
  MODE_VERSION: [
    f'./xmipp {MODE_VERSION}',
    f'./xmipp {MODE_VERSION} {PARAMS[PARAM_SHORT][LONG_VERSION]}',
  ],
  MODE_COMPILE_AND_INSTALL: [
    f'./xmipp {MODE_COMPILE_AND_INSTALL}',
    f'./xmipp {MODE_COMPILE_AND_INSTALL} {PARAMS[PARAM_JOBS][SHORT_VERSION]} 20',
  ],
  MODE_ALL: [
    './xmipp',
    f'./xmipp {MODE_ALL}',
    f'./xmipp {PARAMS[PARAM_JOBS][SHORT_VERSION]} 20',
    f'./xmipp {PARAMS[PARAM_BRANCH][SHORT_VERSION]} devel',
    f'./xmipp {MODE_ALL} {PARAMS[PARAM_JOBS][SHORT_VERSION]} 20 '
    f'{PARAMS[PARAM_BRANCH][SHORT_VERSION]} devel'
  ],
  MODE_CONFIG_BUILD: [],
  MODE_CONFIG: [
    f'./xmipp {MODE_CONFIG} {PARAMS[PARAM_OVERWRITE][LONG_VERSION]}'
  ],
  MODE_GET_MODELS: [
    f'./xmipp {MODE_GET_MODELS}',
    f'./xmipp {MODE_GET_MODELS} {PARAMS[PARAM_MODELS_DIRECTORY][SHORT_VERSION]} /path/to/my/model/directory'
  ],
  MODE_GET_SOURCES: [
    f'./xmipp {MODE_GET_SOURCES}'
    f'./xmipp {MODE_GET_SOURCES} {PARAMS[PARAM_BRANCH][SHORT_VERSION]} devel'
  ],
  MODE_CLEAN_BIN: [],
  MODE_CLEAN_ALL: [],
  MODE_TEST: [
    f'./xmipp {MODE_TEST} xmipp_sample_test',
    f'./xmipp {MODE_TEST} {PARAMS[PARAM_SHOW_TESTS][LONG_VERSION]}',
    f'./xmipp {MODE_TEST} {PARAMS[PARAM_ALL_FUNCTIONS][LONG_VERSION]}',
    f'./xmipp {MODE_TEST} {PARAMS[PARAM_ALL_PROGRAMS][LONG_VERSION]}'
  ],
  MODE_GIT: [
    f'./xmipp {MODE_GIT} pull',
    f'./xmipp {MODE_GIT} checkout devel'
  ],
  MODE_ADD_MODEL: [
    f'./xmipp {MODE_ADD_MODEL} myuser@127.0.0.1 /home/myuser/mymodel'
  ]
}



"""
### Mode CMake Executor Module.

This module contains the base class for executors that run CMake commands.
"""

from abc import abstractmethod
from typing import Dict, Optional, Tuple

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger import errors
from xmipp3_installer.installer.handlers.cmake import cmake_handler
from xmipp3_installer.installer.modes import mode_executor
from xmipp3_installer.repository.config_vars import variables

class ModeCMakeExecutor(mode_executor.ModeExecutor):
  """
  ### Mode CMake Executor.

  Base class for executors that run CMake commands.
  """

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.substitute = not context[params.PARAM_KEEP_OUTPUT]
    self.cmake = context[variables.CMAKE]
    self.build_type = context[variables.BUILD_TYPE]

  def run(self) -> Tuple[int, str]:
    """
    ### Runs the CMake config with the appropiate params.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error. 
    """
    cmake = self.__get_cmake_executable()
    if not cmake:
      return errors.CMAKE_ERROR, "CMake installation not found."
    return self._run_cmake_mode(cmake)
  
  def _set_executor_config(self):
    """### Sets the specific executor params for this mode."""
    super()._set_executor_config()
    self.logs_to_file = True
    self.prints_with_substitution = True
  
  def __get_cmake_executable(self) -> Optional[str]:
    """
    ### Returns the path to the CMake executable to be used.

    #### Returns:
    - (str | None): Path to the CMake executable. None if it was not found.
    """
    return self.cmake or cmake_handler.get_cmake_path()
  
  @staticmethod
  def _get_error_code(ret_code, mode_error_code):
    return ret_code if ret_code == errors.INTERRUPTED_ERROR else mode_error_code
  
  @abstractmethod
  def _run_cmake_mode(self, cmake: str) -> Tuple[int, str]:
    """Run CMake mode method to be implemented by inheriting classes."""



"""
### Mode Get Models Executor Module.

This module contains the class to download deep learning models.
"""

import os
from typing import Dict, Tuple

from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import urls
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler
from xmipp3_installer.installer.modes.mode_sync.mode_sync_executor import ModeSyncExecutor

class ModeGetModelsExecutor(ModeSyncExecutor):
  """
  ### Mode Get Models Executor.

  Downloads deep learning models for the installation.
  """
  
  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.models_directory = context.pop(params.PARAM_MODELS_DIRECTORY)
    if self.models_directory == os.path.abspath(paths.INSTALL_PATH):
      self.models_directory = os.path.join(self.models_directory, 'models')

  def _sync_operation(self) -> Tuple[int, str]:
    """
    ### Downloads deep learning models.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    if os.path.isdir(self.models_directory):
      task = "update"
      in_progress_task = "Updating"
      completed_task = "updated"
    else:
      task = "download"
      in_progress_task = "Downloading"
      completed_task = "downloaded"
    
    logger(f"{in_progress_task} Deep Learning models (in background)")
    ret_code = shell_handler.run_shell_command_in_streaming(
      f"{self.sync_program_path} {task} {self.models_directory} {urls.MODELS_URL} DLmodels",
      show_output=True,
      show_error=True,
      substitute=True
    )
    if not ret_code:
      logger(logger.green(f"Models successfully {completed_task}!"))

    ret_code = 1 if ret_code else ret_code
    return ret_code, ""



"""### Contains all needed CMake constants."""

from xmipp3_installer.repository.config_vars import variables

DEFAULT_CMAKE = 'cmake'

# CMake cache file variables to look for
XMIPP_USE_CUDA=variables.CUDA
XMIPP_USE_MPI=variables.MPI
XMIPP_USE_MATLAB=variables.MATLAB
XMIPP_LINK_TO_SCIPION=variables.LINK_SCIPION
CMAKE_BUILD_TYPE='CMAKE_BUILD_TYPE'
CMAKE_C_COMPILER=variables.CC
CMAKE_CXX_COMPILER=variables.CXX
CMAKE_CUDA_COMPILER=variables.CUDA_COMPILER

# CMake saved version variables
CMAKE_PYTHON = 'Python3'
CMAKE_CUDA = 'CUDA'
CMAKE_MPI = 'MPI'
CMAKE_HDF5 = 'HDF5'
CMAKE_JPEG = 'JPEG'
CMAKE_SQLITE = 'SQLite3'
CMAKE_JAVA = 'Java'
CMAKE_CMAKE = 'CMake'
CMAKE_GCC = 'CC'
CMAKE_GPP = 'CXX'



"""
### Configuration Values Adapter Module.

This module contains functions to convert configuration values between file format and context format.
"""

from typing import Dict, Union

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.repository.config_vars import variables, default_values

def get_context_values_from_file_values(file_values: Dict[str, str], show_warnings: bool=True) -> Dict[str, Union[str, bool]]:
  """
  ### Converts configuration values from file format to context format.
  
  Processes a dictionary of configuration values read from file, converting toggle values 
  from string ('ON'/'OFF') to boolean representation for use in the application context.

  #### Params:
  - file_values (dict(str, str)): Dictionary of configuration values as read from file.
  - show_warnings (bool): Optional. If True, warning messages are shown when applicable.

  #### Returns:
  - (dict(str, str | bool)): Dictionary with values converted to their context format.
  """
  context_values = {}
  for key, value in file_values.items():
    context_values[key] = __get_context_value_from_file_value(key, value, show_warnings)
  return context_values

def get_file_values_from_context_values(context_values: Dict[str, Union[str, bool]]) -> Dict[str, str]:
  """
  ### Converts configuration values from context format to file format.
  
  Processes a dictionary of configuration values from the application context, converting toggle values 
  from boolean to string ('ON'/'OFF') representation for storage in configuration file.

  #### Params:
  - context_values (dict(str, str | bool)): Dictionary of configuration values from context.

  #### Returns:
  - (dict(str, str)): Dictionary with values converted to their file storage format.
  """
  file_values = {}
  for key, value in context_values.items():
    file_values[key] = __get_file_value_from_context_value(key, value)
  return file_values

def __get_context_value_from_file_value(key: str, value: str, show_warnings: bool) -> Union[str, bool]:
  """
  ### Converts a single configuration value from file format to context format.

  #### Params:
  - key (str): Configuration variable key.
  - value (str): Value as read from file.
  - show_warnings (bool): Optional. If True, warning messages are shown when applicable.

  #### Returns:
  - (str | bool): Value converted to its context format.
  """
  if key in variables.CONFIG_VARIABLES[variables.TOGGLES]:
    return __get_boolean_value_from_string(key, value, show_warnings)
  return value

def __get_file_value_from_context_value(key: str, value: Union[str, bool]) -> Union[str, bool]:
  """
  ### Converts a single configuration value from context format to file format.

  #### Params:
  - key (str): Configuration variable key.
  - value (str | bool): Value from context.

  #### Returns:
  - (str): Value converted to its file storage format.
  """
  if key in variables.CONFIG_VARIABLES[variables.TOGGLES]:
    return __get_string_value_from_boolean(bool(value))
  return value

def __get_boolean_value_from_string(key: str, value: str, show_warning: bool) -> bool:
  """
  ### Converts a toggle value from string ('ON'/'OFF') to boolean.

  #### Params:
  - key (str): Configuration variable key.
  - value (str): String value to convert ('ON' or 'OFF').
  - show_warnings (bool): Optional. If True, warning message is shown when applicable.

  #### Returns:
  - (bool): Boolean representation of the toggle value.
  """
  if value not in {default_values.ON, default_values.OFF}:
    default_value = default_values.CONFIG_DEFAULT_VALUES[key]
    if show_warning:
      logger(logger.yellow(
        f"WARNING: config variable '{key}' has unrecognized value '{value}'. "
        f"Toggle values must be either '{default_values.ON}' or '{default_values.OFF}'. "
        f"Default value '{default_value}' will be used instead."
      ))
    value = default_value
  return value == default_values.ON

def __get_string_value_from_boolean(value: bool) -> str:
  """
  ### Converts a toggle value from boolean to string ('ON'/'OFF').

  #### Params:
  - value (bool): Boolean value to convert.

  #### Returns:
  - (str): String representation of the toggle value ('ON' or 'OFF').
  """
  return default_values.ON if value else default_values.OFF



"""### Common formatting functions for parsers."""

from typing import List

from xmipp3_installer.application.cli.arguments import params

TAB_SIZE = 4

def get_formatting_tabs(text: str) -> str:
  """
  ### Returns the given text, formatted to expand tabs into a fixed tab size.

  ### Params:
  - text (str): The text to be formatted.

  ### Returns:
  - (str): Formatted text.
  """
  return text.expandtabs(TAB_SIZE)

def get_param_names(param_key: str) -> List[str]:
  """
  ### Returns the list of possible names a given param has.

  #### Params:
  - param_key (str): Key to find the param.

  #### Returns:
  - (list[str]): Names of the given param.
  """
  names = [
    params.PARAMS[param_key].get(params.SHORT_VERSION, ''),
    params.PARAMS[param_key].get(params.LONG_VERSION, '')
  ]
  return [name for name in names if name]



"""### Contains functions that can interact with packages via shell with a generic interface."""

from typing import Optional

from xmipp3_installer.installer.handlers import shell_handler

def get_package_version(package_name: str) -> Optional[str]:
  """
  ### Retrieves the version of a package or program by executing '[package_name] --version' command.

  Params:
  - package_name (str): Name of the package or program.

  Returns:
  - (str | None): Version information of the package or None if not found or errors happened.
  """
  ret_code, output = shell_handler.run_shell_command(f'{package_name} --version')
  return output if ret_code == 0 else None



"""### Help formatter specific for generic usage mode."""

from xmipp3_installer.application.cli.arguments import modes
from xmipp3_installer.application.cli.parsers import format
from xmipp3_installer.application.cli.parsers.base_help_formatter import BaseHelpFormatter
from xmipp3_installer.application.logger.logger import logger

class GeneralHelpFormatter(BaseHelpFormatter):
  """### Overrides the default help formatter to display a custom help message."""

  def format_help(self):
    """### Prints the help message of the argument parser."""
    help_message = "Run Xmipp's installer script\n\nUsage: xmipp [options]\n"
    for section in modes.MODES.keys():
      help_message += self.__get_section_message(section)
    help_message += f"\n{self.__get_epilog()}"
    help_message += self.__get_note()
    return format.get_formatting_tabs(help_message)
  
  def __get_mode_arg_group_str(self, args: str) -> str:
    """
    ### This method returns the args text for a given arg group.

    ### Params:
    - args (str): Args to format into a message string.

    ### Returns:
    - (str): Args text for given mode.
    """
    param_names = []
    for param in args:
      param_name = self._get_param_first_name(param)
      if param_name:
        param_names.append(f'[{param_name}]')
    return ' '.join(param_names)

  def __get_mode_args_str(self, mode: str) -> str:
    """
    ### This method returns the args text for a given mode.

    ### Params:
    - mode (str): Mode to get args text for.

    ### Returns:
    - (str): Args text for given mode.
    """
    args = modes.MODE_ARGS[mode]
    if not self._has_mutually_exclusive_groups(args):
      return self.__get_mode_arg_group_str(args)
    
    group_strs = [
      self.__get_mode_arg_group_str(arg_group) for arg_group in args
    ]
    exclusive_group_str = " | ".join(group_strs)
    return f"({exclusive_group_str})"

  def __get_mode_args_and_help_str(self, previous_text: str, mode: str) -> str:
    """
    ### This method returns the args and help text for a given mode.

    ### Params:
    - previous_text (str): Text inserted before the one to be returned.
    - mode (str): Mode to get help text for.

    ### Returns:
    - (str): Args and help text for given mode.
    """
    return self._text_with_limits(
      previous_text + self.__get_mode_args_str(mode),
      self._get_mode_help(mode)
    )

  @staticmethod
  def __get_epilog() -> str:
    """
    ### Returns the epilogue.

    #### Returns:
    - (str): Epilogue.
    """
    epilogue = "Example 1: ./xmipp\n"
    epilogue += "Example 2: ./xmipp compileAndInstall -j 4\n"
    return epilogue
  
  @staticmethod
  def __get_note() -> str:
    """
    ### Returns the additional note message.

    #### Returns:
    - (str): Note message.
    """
    note_message = "Note: You can also view a specific help message for each mode with \"./xmipp [mode] -h\".\n"
    note_message += f"Example: ./xmipp {modes.MODE_ALL} -h\n"
    return logger.yellow(note_message)

  def __get_section_message(self, section: str) -> str:
    """
    ### Returns the given section's message.

    #### Params:
    - section (str): Section name.

    #### Return:
    - (str): Section's message.
    """
    section_message = self._get_help_separator() + f"\t# {section} #\n\n"
    for mode in modes.MODES[section].keys():
      section_message += self.__get_mode_args_and_help_str(f"\t{mode} ", mode)
    return section_message



"""
### Mode Clean Bin Executor Module.

This module contains the class to clean compiled binaries and related files.
"""

import fnmatch
import glob
import os
import pathlib
from typing import List

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import constants
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes.mode_clean import mode_clean_executor

class ModeCleanBinExecutor(mode_clean_executor.ModeCleanExecutor):
  """
  ### Mode Clean Bin Executor.

  Cleans compiled binaries and related files.
  """

  @property
  def confirmation_keyword(self) -> str:
    """### Property containineg the keyword to confirm the cleaning operation."""
    return "y"
  
  @staticmethod
  def _get_paths_to_delete() -> List[str]:
    """
    ### Returns a list of all the paths to be deleted.

    #### Returns:
    - (list(str)): List containing all the paths to delete.
    """
    dblite_files = glob.glob(
      "**/*.dblite",
      recursive=True
    )
    return [
      *dblite_files,
      *ModeCleanBinExecutor.__get_compilation_files(),
      *ModeCleanBinExecutor.__get_empty_dirs(),
      *ModeCleanBinExecutor.__get_pycache_dirs(),
      paths.BUILD_PATH,
      paths.BINARIES_PATH
    ]
  
  def _get_confirmation_message(self) -> str:
    """
    ### Returns message to be printed when asking for user confirmation.

    #### Returns:
    - (str): Confirmation message.
    """
    return '\n'.join([
      logger.yellow(f"WARNING: This will DELETE from {paths.SOURCES_PATH} all *.so, *.os and *.o files. Also the *.pyc and *.dblite files"),
      logger.yellow(f"If you are sure you want to do this, type '{self.confirmation_keyword}' (case sensitive):")
    ])
  
  @staticmethod
  def __get_compilation_files():
    """
    ### Returns a list of all the compilation-related files.

    #### Returns:
    - (list(str)): List containing all the paths to compilation-related files.
    """
    compilation_files = []
    for root, _, files in os.walk(paths.SOURCES_PATH):
      for pattern in ['*.so', '*.os', '*.o']:
        for filename in fnmatch.filter(files, pattern):
          compilation_files.append(os.path.join(root, filename))
    return compilation_files

  @staticmethod
  def __get_empty_dirs() -> List[str]:
    """
    ### Returns a list with all the empty directories inside the programs folder.

    #### Returns:
    - (list(str)): List containing the paths to all the empty directories.
    """
    empty_dirs = [] 
    for root, dirs, files in os.walk(os.path.join(
      paths.SOURCES_PATH, constants.XMIPP, "applications", "programs"
    )): 
      if not len(dirs) and not len(files): 
        empty_dirs.append(root) 
    return empty_dirs

  @staticmethod
  def __get_pycache_dirs() -> List[str]:
    """
    ### Returns a list of all the __pycache__ directories.

    #### Returns:
    - (list(str)): List containing all the paths to __pycache__ directories.
    """
    return [
      str(path) for path in pathlib.Path().rglob('__pycache__')
    ]



"""### Functions that interact with Git via shell."""

import os
from typing import Optional, Tuple

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler

def get_current_branch(dir: str='./') -> str:
  """
  ### Returns the current branch of the repository of the given directory or empty string if it is not a repository or a recognizable tag.
  
  #### Params:
  - dir (str): Optional. Directory of the repository to get current branch from. Default is current directory.
  
  #### Returns:
  - (str): The name of the branch, 'HEAD' if it is a tag, or empty string if the given directory is not a repository or a recognizable tag.
  """
  ret_code, branch_name = shell_handler.run_shell_command("git rev-parse --abbrev-ref HEAD", cwd=dir)
  # If there was an error, we are in no branch
  return '' if ret_code else branch_name

def is_tag(dir: str='./') -> bool:
  """
  ### Returns True if the current Xmipp repository is in a tag.
  
  This happens when the current commit matches a tag.

  #### Params:
  - dir (str): Optional. Directory of the repository where the check will happen. Default is current directory.
  
  #### Returns:
  - (bool): True if the repository is a tag. False otherwise.
  """
  ret_code, _ = shell_handler.run_shell_command("git describe --tags --exact-match HEAD", cwd=dir)
  return not ret_code

def is_branch_up_to_date(dir: str='./') -> bool:
  """
  ### Returns True if the current branch is up to date, or False otherwise or if some error happened.
  
  #### Params:
  - dir (str): Optional. Directory of the repository to get current branch from. Default is current directory.
  
  #### Returns:
  - (bool): True if the current branch is up to date, or False otherwise or if some error happened.
  """
  current_branch = get_current_branch(dir=dir)
  if not current_branch:
    return False
  
  ret_code = shell_handler.run_shell_command("git fetch", cwd=dir)[0]
  if ret_code != 0:
    return False

  latest_local_commit = shell_handler.run_shell_command(f"git rev-parse {current_branch}", cwd=dir)[1]
  ret_code, latest_remote_commit = shell_handler.run_shell_command(f"git rev-parse origin/{current_branch}")
  if ret_code != 0:
    return False
  
  return latest_local_commit == latest_remote_commit

def get_current_commit(dir: str="./") -> str:
  """
  ### Returns the current commit short hash of a given repository.

  #### Params:
  - dir (str): Optional. Directory of repository.

  #### Returns:
  - (str): Current commit short hash, or empty string if it is not a repo or there were errors.
  """
  ret_code, output = shell_handler.run_shell_command("git rev-parse --short HEAD", cwd=dir)
  if ret_code or not output:
    return ''
  return output

def get_commit_branch(commit: str, dir: str="./") -> str:
  """
  ### Returns the name of the commit branch. It can be a branch name or a release name.

  #### Params:
  - commit (str): Commit hash.
  - dir (str): Optional. Directory to repository.

  #### Returns:
  - (str): Name of the commit branch or release.
  """
  ret_code, output = shell_handler.run_shell_command(f"git name-rev {commit}", cwd=dir)
  if ret_code or not output:
    return ''
  return output.replace(commit, "").replace(" ", "")

def branch_exists_in_repo(repo_url: str, branch: str) -> bool:
  """
  ### Checks if the given branch exists in the given repository.

  #### Params:
  - repo (str): Repository to check from.
  - branch (str): Name of the branch to check for.

  #### Returns:
  - (bool): True if the branch exists, False otherwise.
  """
  return __ref_exists_in_repo(repo_url, branch, True)

def tag_exists_in_repo(repo_url: str, tag: str) -> bool:
  """
  ### Checks if the given tag exists in the given repository.

  #### Params:
  - repo_url (str): Repository to check from.
  - tag (str): Name of the tag to check for.

  #### Returns:
  - (bool): True if the tag exists, False otherwise.
  """
  return __ref_exists_in_repo(repo_url, tag, False)

def get_clonable_branch(repo_url: str, preferred_branch: str, viable_tag: Optional[str]) -> Optional[str]:
  """
  ### Decides the target to be cloned from a given repository.

  The preferred branch will be selected if exists,
  followed in priority by the viable tag if provided.
  Finally, if no branch could be selected, None is returned,
  meaning that repository's default branch will be used.

  #### Params:
  - repo_url (str): Url of the repositori to be cloned.
  - preferred_branch (str): Preferred branch to clone into.
  - viable_tag (str | None): If exists, it is returned if branch does not.

  #### Returns:
  - (str | None): Name of the branch to clone the repository into, or None if not found.
  """
  if preferred_branch and branch_exists_in_repo(repo_url, preferred_branch):
    return preferred_branch
  if viable_tag and tag_exists_in_repo(repo_url, viable_tag):
    return viable_tag

def execute_git_command_for_source(command: str, source: str) -> Tuple[int, str]:
  """
  ### Executes the git command for a specific source.

  #### Params:
  - command (str): Command to execute on the source.
  - source (str): The source repository name.

  #### Returns:
  - (tuple(int, str)): Tuple containing the return code and output message.
  """
  source_path = paths.get_source_path(source)
  if not os.path.exists(source_path):
    logger(logger.yellow(
      f"WARNING: Source {source} does not exist in path {source_path}. Skipping."
    ))
    return 0, ""
  
  return shell_handler.run_shell_command(
    f"git {command}",
    cwd=source_path,
    show_output=True,
    show_error=True
  )

def __ref_exists_in_repo(repo_url: str, ref: str, is_branch: bool) -> bool:
  """
  ### Checks if a given reference exists in the given repository.

  #### Params:
  - repo_url (str): Repository to check from.
  - ref (str): Reference to check for.
  - is_branch (bool): If True, the reference is a branch. If False, it is a tag.

  #### Returns:
  - (bool): True if the ref exists, False otherwise.
  """
  ref_type = "heads" if is_branch else "tags"
  ret_code, output = shell_handler.run_shell_command(
    f"git ls-remote --{ref_type} {repo_url}.git refs/{ref_type}/{ref}"
  )
  if ret_code:
    return False
  return f"refs/{ref_type}/{ref}" in output



"""
### Mode Add Model Executor Module.

This module contains the class to upload models to the remote server.
"""

import os
import tarfile
from typing import Dict, Tuple

from xmipp3_installer.application import user_interactions
from xmipp3_installer.application.logger import errors
from xmipp3_installer.application.cli.arguments import params
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler
from xmipp3_installer.installer.modes.mode_sync.mode_sync_executor import ModeSyncExecutor

class ModeAddModelExecutor(ModeSyncExecutor):
  """
  ### Mode Add Model Executor.

  Uploads models to the remote server.
  """

  def __init__(self, context: Dict):
    """
    ### Constructor.
    
    #### Params:
    - context (dict): Dictionary containing the installation context variables.
    """
    super().__init__(context)
    self.login = context.pop(params.PARAM_LOGIN)
    self.model_path = context.pop(params.PARAM_MODEL_PATH)
    self.update = context.pop(params.PARAM_UPDATE)
    self.model_dir = os.path.dirname(self.model_path)
    self.model_name = os.path.basename(self.model_path)
    self.tar_file_name = f"xmipp_model_{self.model_name}.tgz"
    self.tar_file_path = os.path.join(self.model_dir, self.tar_file_name)

  def _sync_operation(self) -> Tuple[int, str]:
    """
    ### Uploads the model to the remote server.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    if not os.path.isdir(self.model_path):
      logger('\n'.join([
        logger.red(f"{self.model_path} is not a directory. Please, check the path."),
        logger.red("The name of the model will be the name of that folder.")
      ]))
      return errors.IO_ERROR, ""
    
    ret_code, output = self.__generate_compressed_file()
    if ret_code:
      return ret_code, output
    
    if not self.__get_confirmation():
      return errors.INTERRUPTED_ERROR, ""
    
    ret_code, output = self.__upload_model()
    ret_code = 1 if ret_code else ret_code
    return ret_code, output

  def __generate_compressed_file(self) -> Tuple[int, str]:
    """
    ### Generates the model's compressed file.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    logger(f"Creating the {self.tar_file_name} model.")
    try:
      with tarfile.open(self.tar_file_path, "w:gz") as tar:
        tar.add(self.model_path, arcname=self.model_name)
    except (tarfile.ReadError, tarfile.CompressionError) as ex:
      return errors.IO_ERROR, str(ex)
    return 0, ""

  def __get_confirmation(self) -> bool:
    """
    ### Asks the user for confirmation.

    #### Returns:
    - (bool): True if the user confirms, False otherwise.
    """
    logger('\n'.join([
      logger.yellow("Warning: Uploading, please BE CAREFUL! This can be dangerous."),
      f"You are going to be connected to {self.login} to write in folder {paths.SCIPION_SOFTWARE_EM}.",
      "Continue? YES/no (case sensitive)"
    ]))
    return user_interactions.get_user_confirmation("YES")

  def __upload_model(self) -> Tuple[int, str]:
    """
    ### Uploads the model to the remote server.

    #### Returns:
    - (tuple(int, str)): Tuple containing the return code and an error message if there was an error.
    """
    update = '--update' if self.update else ''
    args = f"{self.login}, {os.path.abspath(self.tar_file_path)}, {paths.SCIPION_SOFTWARE_EM}, {update}"    
    logger(f"Trying to upload the model using {self.login} as login")
    sync_program_relative_call = os.path.join(
      ".",
      os.path.basename(self.sync_program_path)
    )
    ret_code, output = shell_handler.run_shell_command(
      f"{sync_program_relative_call} upload {args}",
      cwd=os.path.dirname(self.sync_program_path)
    )
    if not ret_code:
      output = ""
      logger(logger.green(f"{self.model_name} model successfully uploaded! Removing the local .tgz"))
      os.remove(self.tar_file_path)
    return ret_code, output



"""
### Mode Clean All Executor Module.

This module contains the class to clean all content from source and build directories.
"""

from typing import List

from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer import constants
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.modes.mode_clean import mode_clean_executor

class ModeCleanAllExecutor(mode_clean_executor.ModeCleanExecutor):
  """
  ### Mode Clean All Executor.

  Cleans all content from source and build directories.
  """

  @property
  def confirmation_keyword(self) -> str:
    """### Property containineg the keyword to confirm the cleaning operation."""
    return "YeS"

  @staticmethod
  def _get_paths_to_delete() -> List[str]:
    """
    ### Returns a list of all the paths to be deleted.

    #### Returns:
    - (list(str)): List containing all the paths to delete.
    """
    return [
      *[paths.get_source_path(source) for source in constants.XMIPP_SOURCES],
      paths.INSTALL_PATH,
      paths.BUILD_PATH,
      paths.CONFIG_FILE
    ]

  def _get_confirmation_message(self) -> str:
    """
    ### Returns message to be printed when asking for user confirmation.

    #### Returns:
    - (str): Confirmation message.
    """
    return '\n'.join([
      logger.yellow("WARNING: This will DELETE ALL content from src and build, and also the xmipp.conf file."),
      logger.yellow("\tNotice that if you have unpushed changes, they will be deleted."),
      logger.yellow(f"\nIf you are sure you want to do this, type '{self.confirmation_keyword}' (case sensitive):")
    ])



"""
### Mode Config Build Executor Module.

This module contains the class to configure the build using CMake.
"""

from typing import List, Optional, Tuple, Union, cast

from xmipp3_installer.application.logger import errors, predefined_messages
from xmipp3_installer.application.logger.logger import logger
from xmipp3_installer.installer.constants import paths
from xmipp3_installer.installer.handlers import shell_handler
from xmipp3_installer.installer.handlers.cmake import cmake_handler
from xmipp3_installer.installer.modes.mode_cmake import mode_cmake_executor
from xmipp3_installer.repository.config_vars import variables

class ModeConfigBuildExecutor(mode_cmake_executor.ModeCMakeExecutor):
  """
  ### Mode Config Build Executor.

  Configures the build using CMake with the appropriate parameters.
  """
  
  def _run_cmake_mode(self, cmake: str) -> Tuple[int, str]:
    """
    ### Runs the CMake config with the appropiate params.

    #### Params:
    - cmake (str): Path to CMake executable.

    #### Returns:
    - (tuple(int, str)): Tuple containing the error status and an error message if there was an error. 
    """
    logger(predefined_messages.get_section_message("Configuring with CMake"))
    cmd = f"{cmake} -S . -B {paths.BUILD_PATH} -DCMAKE_BUILD_TYPE={self.build_type} {self.__get_cmake_vars()}"
    ret_code = shell_handler.run_shell_command_in_streaming(cmd, show_output=True, substitute=self.substitute)
    if ret_code:
      return self._get_error_code(ret_code, errors.CMAKE_CONFIGURE_ERROR), ""
    logger(predefined_messages.get_done_message(), substitute=self.substitute)
    return 0, ""
  
  def __get_cmake_vars(self) -> str:
    """
    ### Returns the CMake variables required for the configuration step.

    #### Returns:
    - (str): String containing all required CMake variables
    """
    non_empty_variables = [
      (
        variable_key,
        cast(Union[str, bool], self.context[variable_key])
      ) for variable_key in _get_non_internal_config_vars()
      if not _is_empty(self.context[variable_key])
    ]
    return cmake_handler.get_cmake_params(non_empty_variables)
  
def _get_non_internal_config_vars() -> List[str]:
  """
  ### Returns all non-internal config variable keys.

  #### Returns:
  - (list(str)): A list containing all non-internal config variable keys.
  """
  all_config_var_keys = [
    config_var for variable_section in variables.CONFIG_VARIABLES.values()
    for config_var in variable_section
  ]
  non_internal_keys = list(set(all_config_var_keys) - set(variables.INTERNAL_LOGIC_VARS))
  non_internal_keys.sort() # To keep order consistency
  return non_internal_keys

def _is_empty(value: Optional[Union[bool, str]]) -> bool:
  """
  ### Checks if the given config value is empty.

  #### Params:
  - value (bool | str | None): Value to be checked.

  #### Returns:
  - (bool): True if it is empty, False otherwise.
  """
  return not isinstance(value, bool) and not value



"""### Contains an API client that registers the installation attempts."""

import http.client
import json
from typing import Dict
from urllib.parse import urlparse, ParseResult

from xmipp3_installer.installer import urls
from xmipp3_installer.application.logger.logger import logger

def send_installation_attempt(installation_info: Dict):
  """
  ### Sends a POST request to Xmipp's metrics's API.
  
  #### Params:
  - installation_info (dict): Dictionary containing all the installation information.
  """
  if installation_info is None:
    return
  params = json.dumps(installation_info)
  headers = {"Content-type": "application/json"}
  parsed_url = urlparse(urls.API_URL)
  conn = None
  try:
    conn = __get_https_connection(parsed_url, 6)
    conn.request("POST", parsed_url.path, body=params, headers=headers)
    conn.getresponse()
  except TimeoutError:
    logger(
      logger.yellow("There was a timeout while sending installation data."),
      show_in_terminal=False
    )
  finally:
    if conn is not None:
      conn.close()


def __get_https_connection(parsed_url: ParseResult, timeout_seconds: int) -> http.client.HTTPSConnection:
  """
  ### Establishes the connection needed to send the API call.
  
  Separated to enable integration & E2E tests.

  #### Params:
  - parsed_url (ParseResult): Object containing all elements of the url.
  - timeout_seconds (int): Number of seconds to wait until a timeout.

  #### Returns:
  - (HTTPSConnection): Established connection.
  """
  return http.client.HTTPSConnection(parsed_url.hostname or "localhost", parsed_url.port, timeout=timeout_seconds)



"""### Functions that interact with CMake."""

import os
import shutil
from typing import Any, Dict, Optional, Tuple, Union, List

from xmipp3_installer.installer.handlers.cmake import cmake_constants

def get_cmake_path() -> Optional[str]:
  """
  ### Returns the path to the CMake executable.

  #### Returns:
  - (str | None): Path to the CMake executable if found, None otherwise.
  """
  return shutil.which(cmake_constants.DEFAULT_CMAKE)

def get_cmake_params(variables: List[Tuple[str, Union[str, bool]]]) -> str:
  """
  ### Converts the given list of variable names into CMake parameters.

  #### Params:
  - variables (list): List of variable names to obtain the parameters from.
  """
  result = []
  for key, value in variables:
    result.append(f"-D{key}={value}")
  return ' '.join(result)

def get_library_versions_from_cmake_file(path: str) -> Dict[str, Any]:
  """
  ### Obtains the library versions from the CMake cache file.

  #### Params:
  - path (str): Path to the file containing all versions.

  #### Returns:
  - (dict(str, any)): Dictionary containing all the library versions in the file.
  """
  if not os.path.exists(path):
    return {}
  
  result = {}
  with open(path, encoding="utf-8") as versions_file:
    for line in versions_file.readlines():
      result.update(__get_library_version_from_line(line))
  return result

def __get_library_version_from_line(version_line: str) -> Dict[str, Any]:
  """
  ### Retrieves the name and version of the library in the given line.
  
  #### Params:
  - version_line (str): Text line containing the name and version of the library.
  
  #### Returns:
  - (dict(str, any)): Dictionary where the key is the name and the value is the version.
  """
  TOKEN_NUMBER = 2 # Two tokens separated by a =
  library_with_version = {}
  name_and_version = version_line.replace("\n", "").split('=')
  if len(name_and_version) == TOKEN_NUMBER:
    version = name_and_version[1] if name_and_version[1] else None
    library_with_version[name_and_version[0]] = version
  return library_with_version
